/** @file dibsection.cpp

	HITACHI BASIC MASTER LEVEL3 Mark5 Emulator 'EmuB-6892'
	Skelton for retropc emulator

	@author Sasaji
	@date   2013.10.30 -

	@brief [ dibsection ]
*/
#include "dibsection.h"

DibSection::DibSection()
{
	hMainDib = NULL;
	hMainBuf = NULL;
	hMainBmp = NULL;
	pMainBuf = NULL;

	enable = false;
}

DibSection::~DibSection()
{
	if (enable) {
		ReleaseDib();
	}
}

// DIB쐬
bool DibSection::CreateDib(HDC hdc, long width, long height, HANDLE *hBuf, HBITMAP *hBmp, DWORD **pBuf, HDC *hdcDib)
{
	LPBITMAPINFO bmi;

	if (*hBuf != NULL) {
		ReleaseDib(hBuf, hBmp, hdcDib);
	}

	*hBuf = GlobalAlloc(GPTR, sizeof(BITMAPINFOHEADER) + sizeof(RGBQUAD));
	if (*hBuf == NULL) {
		enable = false;
		return false;
	}

	bmi = (LPBITMAPINFO)(*hBuf);

	memset(&(bmi->bmiHeader), 0, sizeof(BITMAPINFOHEADER));

	bmi->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	bmi->bmiHeader.biWidth = width;
	bmi->bmiHeader.biHeight = -height;	// minus
	bmi->bmiHeader.biPlanes = 1;
	bmi->bmiHeader.biBitCount = 32;
	bmi->bmiHeader.biCompression = 0;
	bmi->bmiHeader.biSizeImage = width * height * 4;

	*hBmp = CreateDIBSection(hdc, bmi, DIB_RGB_COLORS, (VOID **)pBuf, NULL, 0);
	if (*hBmp == NULL) {
		enable = false;
		return false;
	}

	*hdcDib = CreateCompatibleDC(hdc);
	SelectObject(*hdcDib, *hBmp);

	enable = true;
	return true;
}
bool DibSection::CreateDib(HDC hdc, long width, long height)
{
	return CreateDib(hdc, width, height, &hMainBuf, &hMainBmp, &pMainBuf, &hMainDib);
}

// DIB폜
void DibSection::ReleaseDib(HANDLE *hBuf, HBITMAP *hBmp, HDC *hdcDib)
{
	if (*hdcDib) {
		DeleteDC(*hdcDib);
		*hdcDib = NULL;
	}
	if (*hBmp) {
		DeleteObject(*hBmp);
		*hBmp = NULL;
	}
	if (*hBuf) {
		GlobalFree(*hBuf);
		*hBuf = NULL;
	}
	enable = false;
}
void DibSection::ReleaseDib()
{
	ReleaseDib(&hMainBuf, &hMainBmp, &hMainDib);
}