/** @file keybindbox.cpp

	HITACHI BASIC MASTER LEVEL3 Mark5 Emulator 'EmuB-6892'
	Skelton for retropc emulator

	@author Sasaji
	@date   2012.03.31 -

	@brief [ keybind box ]
*/

#include "keybindbox.h"
#include "../emu.h"

KeybindBox::KeybindBox(HINSTANCE hInst, CFont *new_font, EMU *new_emu)
	: CDialogBox(hInst, IDD_KEYBIND, new_font, new_emu)
{

	memset(vmkey_map, 0, sizeof(vmkey_map));

	memset(vkkey_defmap, 0, sizeof(vkkey_defmap));
	memset(vkkey_defmap_rows, 0, sizeof(vkkey_defmap_rows));
	memset(vkkey_defmap_cols, 0, sizeof(vkkey_defmap_cols));

	memset(vkkey_map, 0, sizeof(vkkey_map));
	memset(vkkey_preset_map, 0, sizeof(vkkey_preset_map));
}

void KeybindBox::SetVmKeyMap(int tab_num, UINT16 *vmKeyMap, int size)
{
	if (tab_num < 0 || tab_num >= MAX_TAB_NUM) return;

	for(int i=0; i<KBCTRL_MAX_LINES && i<size; i++) {
		vmkey_map[tab_num][i] = vmKeyMap[i];
	}
	vmkey_map_size[tab_num] = (size > KBCTRL_MAX_LINES) ? KBCTRL_MAX_LINES : size;
}

void KeybindBox::SetVmKey(int tab_num, int idx, UINT16 code)
{
	if (tab_num < 0 || tab_num >= MAX_TAB_NUM) return;

	vmkey_map[tab_num][idx] = code;
}

void KeybindBox::SetVkKeyMap(int tab_num, UINT32 *vkKeyMap)
{
	if (tab_num < 0 || tab_num >= MAX_TAB_NUM) return;

	vkkey_map[tab_num] = vkKeyMap;
}

void KeybindBox::SetVkKeyDefMap(int tab_num, UINT32 *vkKeyDefMap, int rows, int cols)
{
	if (tab_num < 0 || tab_num >= MAX_TAB_NUM) return;

	vkkey_defmap[tab_num] = vkKeyDefMap;
	vkkey_defmap_rows[tab_num] = rows;
	vkkey_defmap_cols[tab_num] = cols;
}

void KeybindBox::SetVkKeyPresetMap(int tab_num, UINT32 *vkKeyMap, int idx)
{
	if (tab_num < 0 || tab_num >= MAX_TAB_NUM) return;

	if (idx >= 0 && idx < MAX_PRESET_NUM) {
		vkkey_preset_map[tab_num][idx] = vkKeyMap;
	}
}

void KeybindBox::SetTitleLabel(_TCHAR *vm, _TCHAR *vk, _TCHAR *vj)
{
	_tcscpy(vm_label, vm);
	_tcscpy(vk_label, vk);
	_tcscpy(vj_label, vj);
}

INT_PTR KeybindBox::onInitDialog(UINT message, WPARAM wParam, LPARAM lParam)
{
	HWND hCtrl;
	TCITEM tcitm;
	UINT32 code;

	CDialogBox::onInitDialog(message, wParam, lParam);

	// get parameters
	int presets = emu->get_parami(EMU::ParamVkKeyPresets);
	presets = (presets < KEYBIND_PRESETS ? presets : KEYBIND_PRESETS);
	for (int i=0; i<MAX_TAB_NUM; i++) {
		SetVmKeyMap(i, (uint16 *)emu->get_paramv(EMU::ParamVmKeyMap0 + i), emu->get_parami(EMU::ParamVmKeyMapSize));

		SetVkKeyDefMap(i, (uint32 *)emu->get_paramv(EMU::ParamVkKeyDefMap0 + i)
			, emu->get_parami(EMU::ParamVkKeyMapKeys), emu->get_parami(EMU::ParamVkKeyMapAssign));

		SetVkKeyMap(i, (uint32 *)emu->get_paramv(EMU::ParamVkKeyMap0 + i));

		for(int j=0; j<presets; j++) {
			SetVkKeyPresetMap(i, (uint32 *)emu->get_paramv(EMU::ParamVkKeyPresetMap00 + i + j * MAX_TAB_NUM), j);
		}
	}
	SetTitleLabel(_T("level3 key"),_T("bind"),_T("joypad"));

	// create dialog
	selected_tabctrl = 0;

	hCtrl = GetDlgItem(hDlg, IDC_CUSTOM1);
	kbctl[0] = KeyBindControl_GetPtr(hCtrl);
	hCtrl = GetDlgItem(hDlg, IDC_CUSTOM2);
	kbctl[1] = KeyBindControl_GetPtr(hCtrl);

	kbctl[0]->Init(1);
	kbctl[1]->Init(2);

	SIZE siz;
	font->GetTextSize(hDlg, NULL, &siz);
	kbctl[0]->SetCellSize(siz.cx * 16 + padding * 2, siz.cy + padding * 2);
	kbctl[1]->SetCellSize(siz.cx * 16 + padding * 2, siz.cy + padding * 2);

	kbctl[0]->MapDefaultVmKeyCode((UINT16 *)vmkey_map[0]);
	kbctl[1]->MapDefaultVmKeyCode((UINT16 *)vmkey_map[1]);

	for(int tab_num=0; tab_num<MAX_TAB_NUM; tab_num++) {
		for(int row=0; row<vmkey_map_size[tab_num]; row++) {
			if (vmkey_map[tab_num][row] < vkkey_defmap_rows[tab_num]) {
				for(int col=0; col<vkkey_defmap_cols[tab_num]; col++) {
					code = *(vkkey_map[tab_num] + vmkey_map[tab_num][row] * vkkey_defmap_cols[tab_num] + col);
					if (tab_num == 1) kbctl[tab_num]->SetVkJoyCode(row, col, code);
					else kbctl[tab_num]->SetVkKeyCode(row, col, code);
				}
			}
		}
		if (tab_num == 1) {
			code = *(vkkey_map[tab_num] + (vkkey_defmap_rows[tab_num] - 1) * vkkey_defmap_cols[tab_num]);
			CheckDlgButton(hDlg, IDC_CHK_COMBI1, code);
		}
	}
	kbctl[0]->SetTitleLabel(vm_label, vk_label);
	kbctl[1]->SetTitleLabel(vm_label, vj_label);


	// ^uRg[
	HWND hTabCtrl = GetDlgItem(hDlg, IDC_TAB1);

	tcitm.mask = TCIF_TEXT;

	tcitm.pszText = _T("keyboard");
	TabCtrl_InsertItem(hTabCtrl , 0 , &tcitm);
	tcitm.pszText = _T("joypad");
	TabCtrl_InsertItem(hTabCtrl , 1 , &tcitm);
	TabCtrl_SetCurSel(hTabCtrl, selected_tabctrl);

	//
	// adjust control size
	//
	RECT re;
	RECT re_tab;
	GetClientRect(hTabCtrl, &re_tab);
	TabCtrl_AdjustRect(hTabCtrl, FALSE, &re_tab);
	re_tab.top += margin + margin;
	re_tab.left += margin + margin;

	// kb control
	re.top = re_tab.top;
	re.left = re_tab.left;
	AdjustControl(IDC_CUSTOM1, &re, kbctl[0]->GetWidth(), 340, false, SM_CXVSCROLL);
	AdjustControl(IDC_CUSTOM2, &re, kbctl[1]->GetWidth(), 340, false, SM_CXVSCROLL);
	int kb_top = re.top;
	int kb_right = re.right;

	// checkbox size
	re.top = re.bottom + padding;
	AdjustCheckBox(IDC_CHK_COMBI1, &re);

	// buttons
	re.bottom = kb_top - margin;
	re.left = kb_right + margin + margin;
	for(int i=IDC_BTN_LOAD_DEFAULT; i<=IDC_BTN_SAVE_PRESET4; i++) {
		re.top = re.bottom + margin;
		if (i == IDC_BTN_LOAD_PRESET1 || i == IDC_BTN_SAVE_PRESET1) {
			re.top += margin;
		}
		AdjustButton(i, 8, &re);
	}

	// tab control size
	re.top = margin;
	re.left = margin;
	AdjustControl(IDC_TAB1, &re, GetRight() + margin, GetBottom() + margin);
	re_tab.top = re_tab.top - margin + 1;
	re_tab.left = re.left + 2;
	re_tab.bottom = re.bottom - re_tab.top - 3;
	re_tab.right = re.right - re_tab.left - 4;
	AdjustControl(IDC_STATIC_0, &re_tab, re_tab.right, re_tab.bottom);

	// ok cancel
	re.top = GetBottom() + margin;
	re.right = GetRight();
	AdjustButton(IDCANCEL, 8, &re, true);
	re.right = re.left - margin;
	AdjustButton(IDOK, 8, &re, true);

	// dialog size
	AdjustWindow(margin);

	select_tabctrl(selected_tabctrl);

	return (INT_PTR)TRUE;
}

INT_PTR KeybindBox::onCommand(UINT message, WPARAM wParam, LPARAM lParam)
{
	switch(LOWORD(wParam)) {
	case IDOK:
	case IDCANCEL:
		if (LOWORD(wParam) == IDOK) {
			onClickOk();
		}
		::EndDialog(hDlg, LOWORD(wParam));
		return (INT_PTR)TRUE;

	case IDC_BTN_LOAD_DEFAULT:
		return onClickLoadDefault();

	case IDC_BTN_LOAD_PRESET1:
	case IDC_BTN_LOAD_PRESET2:
	case IDC_BTN_LOAD_PRESET3:
	case IDC_BTN_LOAD_PRESET4:
		return onClickLoadPreset(LOWORD(wParam) - IDC_BTN_LOAD_PRESET1);

	case IDC_BTN_SAVE_PRESET1:
	case IDC_BTN_SAVE_PRESET2:
	case IDC_BTN_SAVE_PRESET3:
	case IDC_BTN_SAVE_PRESET4:
		return onClickSavePreset(LOWORD(wParam) - IDC_BTN_SAVE_PRESET1);

	}
	return (INT_PTR)FALSE;
}

INT_PTR KeybindBox::onNotify(UINT message, WPARAM wParam, LPARAM lParam)
{
	// ^u؂ւ
	LPNMHDR lpNmHdr = (NMHDR *)lParam;
	int i;
	if (lpNmHdr->idFrom = IDC_TAB1) {
		switch (lpNmHdr->code) {
		case TCN_SELCHANGE:
			i = TabCtrl_GetCurSel(lpNmHdr->hwndFrom);
			select_tabctrl(i);
			break;
		}
	}
	return (INT_PTR)TRUE;
}

INT_PTR KeybindBox::onMouseWheel(UINT message, WPARAM wParam, LPARAM lParam)
{
	if (selected_tabctrl == 1) {
		SendDlgItemMessage(hDlg, IDC_CUSTOM2, message, wParam, lParam);
	} else {
		SendDlgItemMessage(hDlg, IDC_CUSTOM1, message, wParam, lParam);
	}
	return (INT_PTR)TRUE;
}

#if 0
INT_PTR KeybindBox::onControlColorStatic(UINT message, WPARAM wParam, LPARAM lParam)
{
	HANDLE h = (HANDLE)GetStockObject(NULL_BRUSH);
	SetBkMode((HDC)wParam, TRANSPARENT);
	return (INT_PTR)h;
}
#endif

INT_PTR KeybindBox::onClickOk()
{
	UINT32 code;

	for(int tab_num=0; tab_num<MAX_TAB_NUM; tab_num++) {
		for(int row=0; row<vmkey_map_size[tab_num]; row++) {
			if (vmkey_map[tab_num][row] < vkkey_defmap_rows[tab_num]) {
				for(int col=0; col<vkkey_defmap_cols[tab_num]; col++) {
					code = kbctl[tab_num]->GetVkKeyCode(row, col);
					*(vkkey_map[tab_num] + vmkey_map[tab_num][row] * vkkey_defmap_cols[tab_num] + col) = code;
				}
			}
		}
		if (tab_num == 1) {
			*(vkkey_map[tab_num] + (vkkey_defmap_rows[tab_num] - 1) * vkkey_defmap_cols[tab_num])
				= (IsDlgButtonChecked(hDlg, IDC_CHK_COMBI1) ? 1 : 0);
		}
	}
	return (INT_PTR)TRUE;
}

INT_PTR KeybindBox::onClickLoadDefault()
{
	UINT32 code;
	int tab_num = selected_tabctrl;

	for(int i=0; i<vmkey_map_size[tab_num]; i++) {
		if (vmkey_map[tab_num][i] < vkkey_defmap_rows[tab_num]) {
			for(int j=0; j<vkkey_defmap_cols[tab_num]; j++) {
				code = *(vkkey_defmap[tab_num] + vmkey_map[tab_num][i] * vkkey_defmap_cols[tab_num] + j);
				if (tab_num == 1) kbctl[tab_num]->SetVkJoyCode(i, j, code);
				else kbctl[tab_num]->SetVkKeyCode(i, j, code);
			}

		}
	}
	if (tab_num == 1) {
		code = *(vkkey_defmap[tab_num] + (vkkey_defmap_rows[tab_num] - 1) * vkkey_defmap_cols[tab_num]);
		CheckDlgButton(hDlg, IDC_CHK_COMBI1, code);
	}
	return (INT_PTR)TRUE;
}

INT_PTR KeybindBox::onClickLoadPreset(int idx)
{
	UINT32 code;
	int tab_num = selected_tabctrl;

	for(int i=0; i<vmkey_map_size[tab_num]; i++) {
		if (vmkey_map[tab_num][i] < vkkey_defmap_rows[tab_num]) {
			for(int j=0; j<vkkey_defmap_cols[tab_num]; j++) {
				code = *(vkkey_preset_map[tab_num][idx] + vmkey_map[tab_num][i] * vkkey_defmap_cols[tab_num] + j);
				if (tab_num == 1) kbctl[tab_num]->SetVkJoyCode(i, j, code);
				else kbctl[tab_num]->SetVkKeyCode(i, j, code);
			}

		}
	}
	if (tab_num == 1) {
		code = *(vkkey_preset_map[tab_num][idx] + (vkkey_defmap_rows[tab_num] - 1) * vkkey_defmap_cols[tab_num]);
		CheckDlgButton(hDlg, IDC_CHK_COMBI1, code);
	}

	return (INT_PTR)TRUE;
}

INT_PTR KeybindBox::onClickSavePreset(int idx)
{
	UINT32 code;
	int tab_num = selected_tabctrl;

	for(int i=0; i<vmkey_map_size[tab_num]; i++) {
		if (vmkey_map[tab_num][i] < vkkey_defmap_rows[tab_num]) {
			for(int j=0; j<vkkey_defmap_cols[tab_num]; j++) {
				code = kbctl[tab_num]->GetVkKeyCode(i, j);
				*(vkkey_preset_map[tab_num][idx] + vmkey_map[tab_num][i] * vkkey_defmap_cols[tab_num] + j) = code;
			}

		}
	}
	if (tab_num == 1) {
		*(vkkey_preset_map[tab_num][idx] + (vkkey_defmap_rows[tab_num] - 1) * vkkey_defmap_cols[tab_num])
			= (IsDlgButtonChecked(hDlg, IDC_CHK_COMBI1) ? 1 : 0);
	}

	return (INT_PTR)TRUE;
}

void KeybindBox::select_tabctrl(int tab_num)
{
	HWND hCtrl;

	selected_tabctrl = tab_num;
	if (selected_tabctrl == 1) {
		hCtrl = GetDlgItem(hDlg, IDC_CUSTOM1);
		ShowWindow(hCtrl, SW_HIDE);
		hCtrl = GetDlgItem(hDlg, IDC_CUSTOM2);
		ShowWindow(hCtrl, SW_SHOW);
		hCtrl = GetDlgItem(hDlg, IDC_CHK_COMBI1);
		ShowWindow(hCtrl, SW_SHOW);
	} else {
		hCtrl = GetDlgItem(hDlg, IDC_CUSTOM1);
		ShowWindow(hCtrl, SW_SHOW);
		hCtrl = GetDlgItem(hDlg, IDC_CUSTOM2);
		ShowWindow(hCtrl, SW_HIDE);
		hCtrl = GetDlgItem(hDlg, IDC_CHK_COMBI1);
		ShowWindow(hCtrl, SW_HIDE);
	}
}
