/// @file keybindctrl.cpp
///
/// @brief L[R[htRg[
///
/// @author Sasaji
/// @date   2012.3.31
///

#include <stdio.h>
#include "keybindctrl.h"
#include "../res/resource.h"

#define TIMER_JOYPAD 1

ATOM KeyBindControl::RegisterClass(HINSTANCE hInst)
{
	WNDCLASSEX wcex;

	wcex.cbSize = sizeof(WNDCLASSEX);

	wcex.style			= CS_HREDRAW | CS_VREDRAW | CS_DBLCLKS;
	wcex.lpfnWndProc	= KeyBindControl::CustomCtrlProc;
	wcex.cbClsExtra		= 0;
	wcex.cbWndExtra		= 0;
	wcex.hInstance		= hInst;
	wcex.hIcon			= NULL;
	wcex.hCursor		= LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground	= (HBRUSH)(COLOR_WINDOW + 1);
	wcex.lpszMenuName	= NULL;
	wcex.lpszClassName	= TEXT("KeyBindCtrl");
	wcex.hIconSm		= NULL;

	return ::RegisterClassEx(&wcex);
}

void KeyBindControl::UnregisterClass(HINSTANCE hInst)
{
	::UnregisterClass(TEXT("KeyBindCtrl"), hInst);
}

KeyBindControl *KeyBindControl_GetPtr(HWND hWnd)
{
	return (KeyBindControl *)GetWindowLongPtr(hWnd, GWLP_USERDATA);
}

LRESULT CALLBACK KeyBindControl::CustomCtrlProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	HINSTANCE hInst;
	KeyBindControl *kbctl;
	LRESULT re;

	if (message == WM_CREATE) {
		// KeyBindCtrlEBhEKeyBindControl̃CX^X֘AtB
		hInst = (HINSTANCE)GetWindowLongPtr(hWnd, GWLP_HINSTANCE);
		kbctl = new KeyBindControl(hInst, hWnd);
		SetWindowLongPtr(hWnd, GWLP_USERDATA, (LONG_PTR)kbctl);
	} else {
		kbctl = (KeyBindControl *)GetWindowLongPtr(hWnd, GWLP_USERDATA);
	}
	// CX^X̃vV[WĂ
	re = kbctl->Proc(hWnd, message, wParam, lParam);

	if (message == WM_NCDESTROY) {
		// KeyBindCtrlEBhEKeyBindControl̃CX^XjB
		SetWindowLongPtr(hWnd, GWLP_USERDATA, (LONG_PTR)NULL);
		delete kbctl;
	}
	return re;
}
LRESULT CALLBACK KeyBindControl::CustomCtrlMainProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	HWND hMainCtrl = GetParent(hWnd);
	KeyBindControl *kbctl = (KeyBindControl *)GetWindowLongPtr(hMainCtrl, GWLP_USERDATA);

	return kbctl->MainProc(hWnd, message, wParam, lParam);
}
LRESULT CALLBACK KeyBindControl::CustomCtrlEditProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	HWND hMainCtrl = GetParent(GetParent(hWnd));
	KeyBindControl *kbctl = (KeyBindControl *)GetWindowLongPtr(hMainCtrl, GWLP_USERDATA);

	return kbctl->EditProc(hWnd, message, wParam, lParam);
}

KeyBindControl::KeyBindControl(HINSTANCE hInst, HWND hWnd)
{
	hInstance = hInst;
	hMainCtrl = hWnd;

	hStaticMain = NULL;
	memset(table, 0, sizeof(table));

	memset(&vScrMain, 0, sizeof(vScrMain));
	vScrMain.cbSize = sizeof(vScrMain);

	cell_height = 18;
	cell_width = 80;
	margin = 5;
	padding = 2;

	ctrl_width = 90;
	ctrl_height = 30;

	usekey_type = 0;
}

KeyBindControl::~KeyBindControl()
{
}

void KeyBindControl::Init(UINT8 flags)
{
	usekey_type = flags;
	if (flags & 2) {
		// use joypad
		init_joypad(hMainCtrl);
	}
}

void KeyBindControl::init_joypad(HWND hWnd)
{
	// joypad
	JOYINFO joyinfo;
	JOYCAPS joycaps;
	MMRESULT re;
	bool enable = false;

	memset(joyid_map, 0xff, sizeof(joyid_map));

	int i = 0;
	int joy_num = joyGetNumDevs();
	for (int joyid=0; joyid < joy_num && joyid < 16 && i < 2; joyid++) {
//		re = joySetCapture(hWnd, i, 0, TRUE);	//
		re = joyGetPos(joyid, &joyinfo);
		if (re == JOYERR_NOERROR) {
			joyid_map[i] = joyid;
			joyGetDevCaps(joyid, &joycaps, sizeof(JOYCAPS));
			UINT32 x = (joycaps.wXmin + joycaps.wXmax) / 2;
			joy_xmin[i] = (joycaps.wXmin + x) / 2;
			joy_xmax[i] = (joycaps.wXmax + x) / 2;
			UINT32 y = (joycaps.wYmin + joycaps.wYmax) / 2;
			joy_ymin[i] = (joycaps.wYmin + y) / 2;
			joy_ymax[i] = (joycaps.wYmax + y) / 2;

			enable = true;
			i++;
		}
	}
	if (enable) {
		SetTimer(hWnd, TIMER_JOYPAD, 32, NULL);
	} else {
		usekey_type &= ~2;
	}
}

void KeyBindControl::term_joypad(HWND hWnd)
{
	if (usekey_type & 2) {
		KillTimer(hWnd, TIMER_JOYPAD);
	}
//	if (usekey_type & 2) {
//		joyReleaseCapture(JOYSTICKID1);
//	}
}

LRESULT KeyBindControl::Proc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam) {
	PAINTSTRUCT ps;
	HDC hdc;
//	_TCHAR str[1000];
//	_stprintf(str,_T("m:%x w:%d l:%d\n"),message,wParam,lParam);
//	OutputDebugString(str);

	switch (message) {
		case WM_CREATE:
			create_children(hWnd);
			break;
		case WM_DESTROY:
			term_joypad(hWnd);
			destroy_children(hWnd);
			break;
		case WM_SETFONT:
			setfont_children(hWnd, wParam);
			break;
		case WM_CTLCOLORSTATIC:
			return ctlcolor_children(hWnd, wParam, lParam);
			break;
		case WM_PAINT:
			GetScrollInfo(hWnd, SB_VERT, &vScrMain);

			hdc = BeginPaint(hWnd, &ps);
			move_children(hWnd, 0, -vScrMain.nPos);
			EndPaint(hWnd, &ps);

			break;
		// cXN[o[
		case WM_VSCROLL:
			switch(LOWORD(wParam)) {
			case SB_TOP:
				vScrMain.nPos = vScrMain.nMin;
				break;
			case SB_BOTTOM:
				vScrMain.nPos = vScrMain.nMax;
				break;
			case SB_LINEUP:
				vScrMain.nPos -= 10;
				if (vScrMain.nPos < vScrMain.nMin) vScrMain.nPos = vScrMain.nMin;
				break;
			case SB_LINEDOWN:
				vScrMain.nPos += 10;
				if (vScrMain.nPos > vScrMain.nMax) vScrMain.nPos = vScrMain.nMax;
				break;
			case SB_PAGEUP:
				vScrMain.nPos -= vScrMain.nPage;
				if (vScrMain.nPos < vScrMain.nMin) vScrMain.nPos = vScrMain.nMin;
				break;
			case SB_PAGEDOWN:
				vScrMain.nPos += vScrMain.nPage;
				if (vScrMain.nPos > vScrMain.nMax) vScrMain.nPos = vScrMain.nMax;
				break;
			case SB_THUMBTRACK:
				vScrMain.nPos = HIWORD(wParam);
				break;
			case SB_THUMBPOSITION:
				vScrMain.nPos = HIWORD(wParam);
				break;
			}
			SetScrollInfo(hWnd , SB_VERT , &vScrMain , TRUE);

			InvalidateRect(hWnd , NULL , TRUE);
			break;
		// zC[}EX
		case WM_MOUSEWHEEL:
			{
				// zC[񂵂
				int delta_y = GET_WHEEL_DELTA_WPARAM(wParam);
				set_scroll_bar_pos(hWnd, -delta_y);
				InvalidateRect(hWnd , NULL , TRUE);
			}
			return 1;
			break;

		// JOYPAD
//		case MM_JOY1MOVE:
//		case MM_JOY1BUTTONDOWN:
		case WM_TIMER:
			{
				HWND hF = GetFocus();
				HWND hPF = GetParent(hF);

				if (hF != NULL && hPF == hStaticMain) {
					// tH[JXΑM
					SendMessage(hF, message, wParam, lParam);
				}
			}
			return 1;
			break;
		case WM_COMMAND:
			switch(LOWORD(wParam)) {
				case IDOK:
				case IDCANCEL:
				{
					HWND hF = GetFocus();
					HWND hPF = GetParent(hF);

					if (hF == hWnd || hPF == hWnd) {
						return 1;	// tH[JXꍇ
					} else {
						return 0;
					}
				}
				break;
			}
			break;
	}

	return DefWindowProc(hWnd , message , wParam , lParam);
}

LRESULT KeyBindControl::MainProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	WNDPROC origProc = (WNDPROC)GetWindowLongPtr(hWnd, GWLP_USERDATA);

	switch(message) {
		case WM_CTLCOLORSTATIC:
		{
			HWND hCtrl = (HWND)lParam;
			HWND hF = GetFocus();
			HWND hPF = GetParent(hF);

			if (hF == hCtrl && hPF == hWnd) {
				// tH[JXĂRg[̔wiFύX
				HDC hDC = (HDC)wParam;
				HBRUSH hBrush = GetSysColorBrush(COLOR_WINDOW);

				return (LRESULT)hBrush;	// wi̐F
			}
		}
		break;
	}

	return CallWindowProc(origProc, hWnd , message , wParam , lParam);
}

LRESULT KeyBindControl::EditProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	codecols_t *col = (codecols_t *)GetWindowLongPtr(hWnd, GWLP_USERDATA);
	WNDPROC origProc;
//	_TCHAR str[1000];
//	_stprintf(str,_T("m:%x w:%d l:%d\n"),message,wParam,lParam);
//	OutputDebugString(str);

	if (col == NULL) {
		return 0;
	}

	origProc = col->proc;

	switch (message) {
		case WM_DESTROY:
			return 0;
			break;
		case WM_GETDLGCODE:
			{
				LRESULT iRet = CallWindowProc(origProc, hWnd , message , wParam , lParam);
				if (lParam) {
					iRet |= DLGC_WANTALLKEYS;
				}
				return iRet;
			}
			break;
		case WM_LBUTTONDBLCLK:
			{
				if (usekey_type & 1) SetVkKeyCode(col, 0);
				else if (usekey_type & 2) SetVkJoyCode(col, 0);
			}
			break;
		case WM_SYSCHAR:
		case WM_CHAR:
			return 0;
			break;
		case WM_SYSKEYDOWN:
		case WM_KEYDOWN:
			{
				if (usekey_type & 1) SetVkKeyCode(col, (UINT16)wParam);
			}
			return 0;
			break;
		case WM_SYSKEYUP:
		case WM_KEYUP:
			return 0;
			break;
//		case MM_JOY1MOVE:
//		case MM_JOY1BUTTONDOWN:
		case WM_TIMER:
			if (usekey_type & 2) {
				JOYINFOEX joyinfoex;
				joyinfoex.dwSize = sizeof(JOYINFOEX);
				joyinfoex.dwFlags = JOY_RETURNALL;
				int i = col->col;
				UINT32 joy_status = 0;
				if(joyid_map[i] != -1 && joyGetPosEx(joyid_map[i], &joyinfoex) == JOYERR_NOERROR) {
					if(joyinfoex.dwYpos < joy_ymin[i]) joy_status |= 0x01;
					if(joyinfoex.dwYpos > joy_ymax[i]) joy_status |= 0x02;
					if(joyinfoex.dwXpos < joy_xmin[i]) joy_status |= 0x04;
					if(joyinfoex.dwXpos > joy_xmax[i]) joy_status |= 0x08;
					if (joy_status == 0) {
						joy_status |= ((joyinfoex.dwButtons & 0x0fffffff) << 4);
					}
					if (joy_status) SetVkJoyCode(col, joy_status);
				}
			}
			return 0;
			break;
	}
	return CallWindowProc(origProc, hWnd , message , wParam , lParam);
}

/// qRg[쐬
///
///
void KeyBindControl::create_children(HWND hWnd)
{
	_TCHAR label[100];
	int y = margin;
	int x = margin;

	vScrMain.cbSize = sizeof(SCROLLINFO);

	memset(table, 0, sizeof(table));

	hStaticMain = CreateWindow(_T("STATIC"), _T(""), WS_CHILD | WS_VISIBLE,	 0, 0, 1, 1, hWnd, NULL, hInstance, NULL);
	if (hStaticMain == NULL) {
		return;
	}
	WNDPROC proc = (WNDPROC)GetWindowLongPtr(hStaticMain, GWLP_WNDPROC);
	SetWindowLongPtr(hStaticMain, GWLP_WNDPROC, (LONG_PTR)CustomCtrlMainProc);
	SetWindowLongPtr(hStaticMain, GWLP_USERDATA, (LONG_PTR)proc);

	// title
	hTitles[0] = CreateWindow(_T("STATIC"), _T("vm key"), WS_CHILD | WS_VISIBLE | SS_CENTER | SS_CENTERIMAGE,
				 x, y, cell_width, cell_height, hStaticMain, NULL, hInstance, NULL);
	x += cell_width + padding;
	for(int j=1; j<=KBCTRL_MAX_COLS; j++) {
		_stprintf(label, _T("bind%d"), j);
		hTitles[j] = CreateWindow(_T("STATIC"), label, WS_CHILD | WS_VISIBLE | SS_CENTER | SS_CENTERIMAGE,
				 x, y, cell_width, cell_height, hStaticMain, NULL, hInstance, NULL);
		x += cell_width + padding;
	}
	y += cell_height + padding;

	for(int i=0; i<KBCTRL_MAX_LINES; i++) {
		x = margin;

		table[i].row = i;

		get_vmkeylabel(i, label);

		table[i].hSta1 = CreateWindow(_T("STATIC"), label, WS_CHILD | WS_VISIBLE | SS_CENTER | SS_CENTERIMAGE,
				 x, y, cell_width, cell_height, hStaticMain, NULL, hInstance, NULL);
		x += cell_width + padding;

		for(int j=0; j<KBCTRL_MAX_COLS; j++) {
			table[i].cols[j].col = j;
			table[i].cols[j].hEdt = CreateWindow(_T("edit"),_T(""),WS_CHILD | WS_VISIBLE | WS_BORDER | ES_CENTER | ES_READONLY,
					 x, y, cell_width, cell_height, hStaticMain, NULL, hInstance, NULL);
			table[i].cols[j].proc = (WNDPROC)GetWindowLongPtr(table[i].cols[j].hEdt, GWLP_WNDPROC);
			SetWindowLongPtr(table[i].cols[j].hEdt, GWLP_WNDPROC, (LONG_PTR)CustomCtrlEditProc);
			SetWindowLongPtr(table[i].cols[j].hEdt, GWLP_USERDATA, (LONG_PTR)&table[i].cols[j]);
			x += cell_width + padding;
		}

		table[i].enabled = true;


		y += cell_height + padding;
	}

	x += margin - padding;
	y += margin - padding;
	ctrl_width = x;
	ctrl_height = y;
	MoveWindow(hStaticMain, 0, 0, x, y, false);

	// XN[o[ݒ
	set_scroll_bar_range(hWnd, y);
}

/// qRg[̃tHgZbg
///
///
void KeyBindControl::setfont_children(HWND hWnd, WPARAM wParam)
{
	HDC hdc = GetDC(hWnd);
	SelectObject(hdc, (HFONT)wParam);
	SIZE siz;
	GetTextExtentPoint32(hdc, _T("0"), 1, &siz);
	ReleaseDC(hWnd, hdc);
	if (siz.cx > 0) cell_width = siz.cx * 10 + padding * 2;
	if (siz.cy > 0) cell_height = siz.cy + padding * 2;

	for(int j=0; j<=KBCTRL_MAX_COLS; j++) {
		if (hTitles[j]) SendMessage(hTitles[j],WM_SETFONT,wParam,MAKELPARAM(TRUE, 0));
	}
	for(int i=0; i<KBCTRL_MAX_LINES; i++) {
		if (table[i].hSta1) SendMessage(table[i].hSta1,WM_SETFONT,wParam,MAKELPARAM(TRUE, 0));
		for(int j=0; j<KBCTRL_MAX_COLS; j++) {
			if (table[i].cols[j].hEdt) SendMessage(table[i].cols[j].hEdt,WM_SETFONT,wParam,MAKELPARAM(TRUE, 0));
		}
	}
}
/// qRg[폜
///
///
void KeyBindControl::destroy_children(HWND hWnd)
{
	for(int i=0; i<KBCTRL_MAX_LINES; i++) {
		for(int j=0; j<KBCTRL_MAX_COLS; j++) {
			SetWindowLongPtr(table[i].cols[j].hEdt, GWLP_WNDPROC, (LONG_PTR)table[i].cols[j].proc);
		}
	}
	WNDPROC proc = (WNDPROC)GetWindowLongPtr(hStaticMain, GWLP_USERDATA);
	SetWindowLongPtr(hStaticMain, GWLP_WNDPROC, (LONG_PTR)proc);
}

/// qRg[zu
///
///
void KeyBindControl::dispose_children(HWND hWnd)
{
	int x = margin;
	int y = margin;

	// title
	for(int j=0; j<=KBCTRL_MAX_COLS; j++) {
		MoveWindow(hTitles[j], x, y, cell_width, cell_height, false);
		x += cell_width + padding;
	}

	y += padding + cell_height + padding;
	x = margin;
	for(int i=0; i<KBCTRL_MAX_LINES; i++) {
		if (table[i].enabled != true) {
			ShowWindow(table[i].hSta1, SW_HIDE);
			for(int j=0; j<KBCTRL_MAX_COLS; j++) {
				ShowWindow(table[i].cols[j].hEdt, SW_HIDE);
			}
		} else {
			x = margin;

			MoveWindow(table[i].hSta1, x, y, cell_width, cell_height, false);
			ShowWindow(table[i].hSta1, SW_SHOW);
			x += cell_width + padding;

			for(int j=0; j<KBCTRL_MAX_COLS; j++) {
				MoveWindow(table[i].cols[j].hEdt, x, y, cell_width, cell_height, false);
				ShowWindow(table[i].cols[j].hEdt, SW_SHOW);
				x += cell_width + padding;
			}
			y += cell_height + padding;
		}
	}

	x += margin - padding;
	y += margin - padding;
	MoveWindow(hStaticMain, 0, 0, x, y, true);

	ctrl_width = x;
	ctrl_height = y;

	// XN[o[ݒ
	set_scroll_bar_range(hWnd, y);
}
//
//
//
LRESULT KeyBindControl::ctlcolor_children(HWND hWnd, WPARAM wParam, LPARAM lParam)
{
	HDC hDC = (HDC)wParam;
	HWND hCtrl = (HWND)lParam;
	HBRUSH hBrush = GetSysColorBrush(COLOR_WINDOW);

	return (LRESULT)hBrush;	// eLXgĂȂ̔wi̐F
}

//
//
//
void KeyBindControl::move_children(HWND hWnd, int x, int y)
{
	WINDOWINFO wid;
	RECT re;

	GetWindowInfo(hStaticMain, &wid);

	re.left = x;
	re.top = y;
	re.right = wid.rcClient.right - wid.rcClient.left;
	re.bottom = wid.rcClient.bottom - wid.rcClient.top;
	MoveWindow(hStaticMain, re.left, re.top, re.right, re.bottom, true);


}

bool KeyBindControl::get_vmkeylabel(int code, _TCHAR *label) {
	bool known_key = true;

	switch(code) {
		case 0x00:
			_tcscpy(label, _T(""));
			break;
		case 0x01:
			_tcscpy(label, _T("break"));
			break;
		case 0x02:
			_tcscpy(label, _T("reset"));
			break;
		case 0x03:
			_tcscpy(label, _T("mode"));
			break;
		case 0x04:
			_tcscpy(label, _T("(Pause)"));
			break;
		case 0x05:
			_tcscpy(label, _T("(Alt)"));
			break;
		case 0x06:
			_tcscpy(label, _T("ins/del"));
			break;
		case 0x07:
			_tcscpy(label, _T("num ins/del"));
			break;
		case 0x08:
			_tcscpy(label, _T("bs"));
			break;
		case 0x09:
			_tcscpy(label, _T("tab"));
			break;
		case 0x0c:
			_tcscpy(label, _T("home/cls"));
			break;
		case 0x0d:
			_tcscpy(label, _T("return"));
			break;
		case 0x0e:
			_tcscpy(label, _T("enter"));
			break;
		case 0x0f:
			_tcscpy(label, _T("help"));
			break;

		case 0x10:
			_tcscpy(label, _T("shift"));
			break;
		case 0x11:
			_tcscpy(label, _T("ctrl"));
			break;
		case 0x14:
			_tcscpy(label, _T("caps"));
			break;
		case 0x15:
			_tcscpy(label, _T("kana"));
			break;
		case 0x16:
			_tcscpy(label, _T("graph"));
			break;
		case 0x17:
			_tcscpy(label, _T("henkan"));
			break;
		case 0x18:
			_tcscpy(label, _T("muhenkan"));
			break;
		case 0x19:
			_tcscpy(label, _T("command"));
			break;
		case 0x1a:
			_tcscpy(label, _T("option"));
			break;

		case 0x1b:
			_tcscpy(label, _T("esc"));
			break;
		case 0x1c:
			_tcscpy(label, _T("right"));
			break;
		case 0x1d:
			_tcscpy(label, _T("left"));
			break;
		case 0x1e:
			_tcscpy(label, _T("up"));
			break;
		case 0x1f:
			_tcscpy(label, _T("down"));
			break;
		case 0x20:
			_tcscpy(label, _T("space"));
			break;
		// 0x21 - 0x28  button1 - 8

		// 0x29 - 0x2f  unknown

		// 0x30 - 0x39  0 - 9
		case 0x3a:
			_tcscpy(label, _T(": *"));
			break;
		case 0x3b:
			_tcscpy(label, _T("; +"));
			break;
		case 0x3c:
			_tcscpy(label, _T(", <"));
			break;
		case 0x3d:
			_tcscpy(label, _T("- ="));
			break;
		case 0x3e:
			_tcscpy(label, _T(". >"));
			break;
		case 0x3f:
			_tcscpy(label, _T("/ ?"));
			break;
		case 0x40:
			_tcscpy(label, _T("@"));
			break;
		// 0x41 - 0x5a  A - Z
		case 0x5b:
			_tcscpy(label, _T("["));
			break;
		case 0x5c:
			_tcscpy(label, _T("\\"));
			break;
		case 0x5d:
			_tcscpy(label, _T("]"));
			break;
		case 0x5e:
			_tcscpy(label, _T("^"));
			break;
		case 0x5f:
			_tcscpy(label, _T("_"));
			break;
		// 0x60 - 0x69  num 0 - num 9
		case 0x6a:
			_tcscpy(label, _T("num *"));
			break;
		case 0x6b:
			_tcscpy(label, _T("num +"));
			break;
		case 0x6c:
			_tcscpy(label, _T("num ?"));
			break;
		case 0x6d:
			_tcscpy(label, _T("num -"));
			break;
		case 0x6e:
			_tcscpy(label, _T("num ."));
			break;
		case 0x6f:
			_tcscpy(label, _T("num /"));
			break;
		// 0x70 - 0x7f  PF1 - PF16

		default:
			if (0x30 <= code && code <= 0x39) {
				// 0 - 9
				_stprintf(label,_T("%c"),code);
			} else if (0x41 <= code && code <= 0x5a) {
				// A - Z
				_stprintf(label,_T("%c"),code);
			} else if (0x60 <= code && code <= 0x69) {
				// num 0 - num 9
				_stprintf(label,_T("num %c"),code - 0x30);
			} else if (0x70 <= code && code <= 0x7f) {
				// PF1 - PF16
				_stprintf(label,_T("PF%d"),code - 0x6f);
			} else {
				_stprintf(label,_T("0x%02x"),code);
				known_key = false;

			}
			break;
	}

	return known_key;
}

UINT32 KeyBindControl::translate_vkkey(UINT32 code) {
	int new_code = code;

	// translate code
	switch(code) {
		case 0xf0:	// caps lock
			new_code = VK_CAPITAL;
			break;
		case 0xf2:	// katakana
			new_code = VK_KANA;
			break;
		case 0xf3:	// kanji
		case 0xf4:	// kanji
			new_code = VK_KANJI;
			break;
	}
	return new_code;
}

bool KeyBindControl::get_vkkeylabel(UINT32 code, _TCHAR *label) {
	bool known_key = true;

	switch(code) {
		case 0:
			_tcscpy(label, _T(""));
			break;
		case VK_BACK:
			_tcscpy(label, _T("bs"));
			break;
		case VK_TAB:
			_tcscpy(label, _T("tab"));
			break;
		case VK_CLEAR:
			_tcscpy(label, _T("clear"));
			break;
		case VK_RETURN:
			_tcscpy(label, _T("enter"));
			break;
		case VK_SHIFT:
			_tcscpy(label, _T("shift"));
			break;
		case VK_CONTROL:
			_tcscpy(label, _T("ctrl"));
			break;
		case VK_MENU:
			_tcscpy(label, _T("alt"));
			break;
		case VK_PAUSE:
			_tcscpy(label, _T("pause"));
			break;
		case VK_CAPITAL:
			_tcscpy(label, _T("caps"));
			break;
		case VK_KANA:
			_tcscpy(label, _T("katakana"));
			break;
		case VK_KANJI:
			_tcscpy(label, _T("kanji"));
			break;
		case VK_ESCAPE:
			_tcscpy(label, _T("esc"));
			break;
		case VK_CONVERT:
			_tcscpy(label, _T("henkan"));
			break;
		case VK_NONCONVERT:
			_tcscpy(label, _T("muhenkan"));
			break;
		case VK_SPACE:
			_tcscpy(label, _T("space"));
			break;
		case VK_PRIOR:
			_tcscpy(label, _T("page up"));
			break;
		case VK_NEXT:
			_tcscpy(label, _T("page down"));
			break;
		case VK_END:
			_tcscpy(label, _T("end"));
			break;
		case VK_HOME:
			_tcscpy(label, _T("home"));
			break;
		case VK_LEFT:
			_tcscpy(label, _T("left"));
			break;
		case VK_UP:
			_tcscpy(label, _T("up"));
			break;
		case VK_RIGHT:
			_tcscpy(label, _T("right"));
			break;
		case VK_DOWN:
			_tcscpy(label, _T("down"));
			break;
		case VK_INSERT:
			_tcscpy(label, _T("insert"));
			break;
		case VK_DELETE:
			_tcscpy(label, _T("delete"));
			break;
		case VK_HELP:
			_tcscpy(label, _T("help"));
			break;
		case VK_LWIN:
			_tcscpy(label, _T("left win"));
			break;
		case VK_RWIN:
			_tcscpy(label, _T("right win"));
			break;
		case VK_APPS:
			_tcscpy(label, _T("menu"));
			break;
		case VK_MULTIPLY:
			_tcscpy(label, _T("num *"));
			break;
		case VK_ADD:
			_tcscpy(label, _T("num +"));
			break;
		case VK_SUBTRACT:
			_tcscpy(label, _T("num -"));
			break;
		case VK_DECIMAL:
			_tcscpy(label, _T("num ."));
			break;
		case VK_DIVIDE:
			_tcscpy(label, _T("num /"));
			break;
		case VK_NUMLOCK:
			_tcscpy(label, _T("numlock"));
			break;
		case VK_SCROLL:
			_tcscpy(label, _T("scroll lock"));
			break;
		case VK_OEM_1:
			_tcscpy(label, _T(": *"));	// JP
			break;
		case VK_OEM_PLUS:
			_tcscpy(label, _T("; +"));	// JP
			break;
		case VK_OEM_COMMA:
			_tcscpy(label, _T(", <"));	// JP
			break;
		case VK_OEM_MINUS:
			_tcscpy(label, _T("- ="));	// JP
			break;
		case VK_OEM_PERIOD:
			_tcscpy(label, _T(". >"));	// JP
			break;
		case VK_OEM_2:
			_tcscpy(label, _T("/ ?"));	// JP
			break;
		case VK_OEM_3:
			_tcscpy(label, _T("@"));	// JP
			break;
		case VK_OEM_4:
			_tcscpy(label, _T("["));	// JP
			break;
		case VK_OEM_5:
			_tcscpy(label, _T("\\"));	// JP
			break;
		case VK_OEM_6:
			_tcscpy(label, _T("]"));	// JP
			break;
		case VK_OEM_7:
			_tcscpy(label, _T("^"));	// JP
			break;
		case VK_OEM_102:
			_tcscpy(label, _T("_"));	// JP
			break;
		default:
			if (0x30 <= code && code <= 0x39) {
				// 0 - 9
				_stprintf(label,_T("%c"),code);
			} else if (0x41 <= code && code <= 0x5a) {
				// A - Z
				_stprintf(label,_T("%c"),code);
			} else if (0x60 <= code && code <= 0x69) {
				// num 0 - num 9
				_stprintf(label,_T("num %c"),code - 0x30);
			} else if (0x70 <= code && code <= 0x87) {
				// F1 - F24
				_stprintf(label,_T("F%d"),code - 0x6f);
			} else {
				_stprintf(label,_T("0x%02x"),code);
				known_key = false;
			}
			break;
	}

	return known_key;
}

bool KeyBindControl::get_vkjoylabel(UINT32 code, _TCHAR *label) {
	bool known_key = true;
	_TCHAR numstr[8];

	switch(code) {
		case 0x00:
			_tcscpy(label, _T(""));
			break;
		case 0x01:
			_tcscpy(label, _T("up"));
			break;
		case 0x02:
			_tcscpy(label, _T("down"));
			break;
		case 0x04:
			_tcscpy(label, _T("left"));
			break;
		case 0x05:
			_tcscpy(label, _T("up+left"));
			break;
		case 0x06:
			_tcscpy(label, _T("down+left"));
			break;
		case 0x08:
			_tcscpy(label, _T("right"));
			break;
		case 0x09:
			_tcscpy(label, _T("up+right"));
			break;
		case 0x0a:
			_tcscpy(label, _T("down+right"));
			break;
		default:
			// button
			if (code & 0xfffffff0) {
				_tcscpy(label, _T("button"));
				for(int i=0; i<28; i++) {
					if (code & (0x10 << i)) {
						_stprintf(numstr, _T(" %d"), i+1);
						_tcscat(label, numstr);
					}
				}
			} else {
				_tcscpy(label, _T(""));
				known_key = false;
			}
			break;
	}

	return known_key;
}

void KeyBindControl::set_scroll_bar_range(HWND hWnd, int y)
{
	RECT rt;
	GetClientRect(hWnd, &rt);

	vScrMain.fMask = SIF_DISABLENOSCROLL | SIF_PAGE | SIF_RANGE;
	vScrMain.nMin = 0;
	vScrMain.nMax = y;
	vScrMain.nPage = rt.bottom - rt.top;
	SetScrollInfo(hWnd , SB_VERT , &vScrMain , TRUE);
	vScrMain.fMask = SIF_POS;
}

/// XN[o[̈ʒuݒ
void KeyBindControl::set_scroll_bar_pos(HWND hWnd, int dy)
{
	vScrMain.fMask = SIF_POS;
	vScrMain.nPos += dy;
	if (vScrMain.nPos > vScrMain.nMax) vScrMain.nPos = vScrMain.nMax;
	if (vScrMain.nPos < vScrMain.nMin) vScrMain.nPos = vScrMain.nMin;
	SetScrollInfo(hWnd , SB_VERT , &vScrMain , TRUE);
}

void KeyBindControl::SetTitleLabel(const _TCHAR *vmlabel, const _TCHAR *vklabel)
{
	_TCHAR label[100];

	SetWindowText(hTitles[0], vmlabel);
	for(int j=1; j<=KBCTRL_MAX_COLS; j++) {
		_stprintf(label, _T("%s%d"), vklabel, j);
		SetWindowText(hTitles[j], label);
	}
}

void KeyBindControl::SetCellSize(int w, int h)
{
	cell_width = w;
	cell_height = h;
}

bool KeyBindControl::SetVmKeyCode(int idx, UINT16 code)
{
	if (idx < 0 || idx >= KBCTRL_MAX_LINES) {
		return false;
	}
	if (code >= 0xff) {
		table[idx].enabled = false;
	} else {
		table[idx].enabled = true;
	}

	table[idx].vm_keycode = code;

	return true;
}

UINT16 KeyBindControl::GetVmKeyCode(int idx)
{
	return table[idx].vm_keycode;
}

bool KeyBindControl::SetVkKeyCode(int idx, int col, UINT32 code)
{
	if (idx < 0 || idx >= KBCTRL_MAX_LINES) {
		return false;
	}
	if (col < 0 || col >= KBCTRL_MAX_COLS) {
		return false;
	}
	if (table[idx].enabled != true) {
		return false;
	}

	return SetVkKeyCode(&table[idx].cols[col], code);
}

bool KeyBindControl::SetVkKeyCode(codecols_t *obj, UINT32 code)
{
	_TCHAR label[100];
	UINT32 new_code = 0;

	new_code = translate_vkkey(code);
	get_vkkeylabel(new_code, label);
	obj->vk_prev_keycode = obj->vk_keycode;
	obj->vk_keycode = new_code;
	SetWindowText(obj->hEdt, label);

	return true;
}

bool KeyBindControl::SetVkJoyCode(int idx, int col, UINT32 code)
{
	if (idx < 0 || idx >= KBCTRL_MAX_LINES) {
		return false;
	}
	if (col < 0 || col >= KBCTRL_MAX_COLS) {
		return false;
	}
	if (table[idx].enabled != true) {
		return false;
	}

	return SetVkJoyCode(&table[idx].cols[col], code);
}

bool KeyBindControl::SetVkJoyCode(codecols_t *obj, UINT32 code)
{
	_TCHAR label[100];

	if (get_vkjoylabel(code, label) == true) {
		obj->vk_prev_keycode = obj->vk_keycode;
		obj->vk_keycode = code;
		SetWindowText(obj->hEdt, label);
	}
	return true;
}

UINT32 KeyBindControl::GetVkKeyCode(int idx, int col)
{
	return table[idx].cols[col].vk_keycode;
}

bool KeyBindControl::UndoVkKeyCode(codecols_t *obj)
{
	return SetVkKeyCode(obj, obj->vk_prev_keycode);
}

bool KeyBindControl::MapDefaultVmKeyCode(UINT16 *tbl)
{
	for(int i=0; i<KBCTRL_MAX_LINES; i++) {
		SetVmKeyCode(i, tbl[i]);
	}
	dispose_children(hMainCtrl);
	return true;
}

