/** @file ledbox.h

	HITACHI BASIC MASTER LEVEL3 Mark5 Emulator 'EmuB-6892'
	Skelton for retropc emulator

	@author Sasaji
	@date   2011.08.08 -

	@brief [ led box ]
*/

#ifndef _LEDBOX_H_
#define _LEDBOX_H_

#include <windows.h>

#include <d3d9.h>
#pragma comment(lib, "d3d9.lib")
#ifdef USE_DIRECT3DX
#include <d3dx9.h>
#pragma comment(lib, "d3dx9.lib")
#endif

#include "dibsection.h"

/**
	@brief LedBox is the class that display the access indicator on the screen.
*/
class LedBox : public DibSection
{
private:
	HWND		hLedBox;
	HINSTANCE	hInstance;
	int			flag_old;

	enum enum_led_type {
		LED_TYPE_BASE = 0,
		LED_TYPE_HR,
		LED_TYPE_HG,
		LED_TYPE_VR,
		LED_TYPE_VG,
		LED_TYPE_RR,
		LED_TYPE_CL,
		LED_TYPE_DIGIT,
		LED_TYPE_END
	};

	// LED
	typedef struct led_st {
		HBITMAP     hBmp;
		HDC         hDC;
		SIZE        sz;
	} led_t;
	led_t       led[LED_TYPE_END];

	enum enum_led_pos {
		LED_POS_BASE = 0,
		LED_POS_KATAKANA,
		LED_POS_HIRAGANA,
		LED_POS_CAPS,
		LED_POS_PLAY,
		LED_POS_REC,
		LED_POS_FDD1,
		LED_POS_FDD2,
		LED_POS_DIGIT1,
		LED_POS_DIGIT2,
		LED_POS_DIGIT3,
		LED_POS_END
	};
	// LEDʒu
	typedef struct led_pt_st {
		POINT pt;
		SIZE  sz;
	} led_pt_t;
	led_pt_t	led_pt[LED_POS_END];

	// \ʒu
	struct win_pt_st {
		int left;
		int top;
		int right;
		int bottom;
		int place;
	} win_pt;
	POINT       parent_pt;
	SIZE        win_sz;

	int			mode;
//	POINT		dist_def[2];
	POINT		dist_set[2];
	POINT		dist;

	bool        visible;
	bool        inside;

	bool create_bitmap(HDC, led_t *, DWORD);
	void copy_led(led_pt_t &, led_t &);
	void copy_led_s(led_pt_t &, led_t &, POINT &);
	void mouse_move(HWND, const POINT &, const LPARAM &);
	void adjust_dialog_size(HWND);

	static INT_PTR CALLBACK LedBoxProc(HWND, UINT, WPARAM, LPARAM);
public:
	LedBox(HINSTANCE);
	~LedBox();
	void InitScreen(HDC, int, int);
	void Create(HWND);
	void Show(int);
	void Move();
	void MoveInPlace();
	void Update(int);
	void SetMode(int);
	void SetPos(int, int, int, int, int = -1);
	void SetPos(int = -1);
	void Draw(HDC);
	void Draw(LPDIRECT3DSURFACE9);
};

#endif
