/** @file logging.cpp

	Skelton for retropc emulator

	@author Sasaji
	@date   2015.02.18 -

	@brief [ logging ]
*/

#include <windows.h>
#include "logging.h"
#include "vm/vm.h"
#include "utility.h"

Logging::Logging()
{
	fplog = NULL;
}

Logging::~Logging()
{
}

void Logging::open_logfile(const _TCHAR *path)
{
#ifdef _LOG_CONSOLE
	AllocConsole();
	hConsole = GetStdHandle(STD_OUTPUT_HANDLE);
	SetConsoleTitle("Debug Log");
#endif
#ifdef _LOG_FILE
	// get application path
	_TCHAR log_path[_MAX_PATH];
	_stprintf(log_path, _T("%s%s.log"), UTILITY::conv_from_native_path(path), _T(CONFIG_NAME));

	fplog = _tfopen(log_path, _T("w"));
#endif
}

void Logging::close_logfile()
{
#ifdef _LOG_CONSOLE
	FreeConsole();
#endif
#ifdef _LOG_FILE
	if(fplog) {
		fclose(fplog);
		fplog = NULL;
	}
#endif
}

void Logging::out_logv(int level, const _TCHAR* format, va_list ap)
{
	_TCHAR buffer[1024];
	_TCHAR levelstr[32];

	_vstprintf(buffer, format, ap);

	switch(level) {
	case LOG_WARN:
		_tcscpy(levelstr, _T("WARN"));
		break;
	case LOG_ERROR:
		_tcscpy(levelstr, _T("ERROR"));
		break;
	case LOG_DEBUG:
		_tcscpy(levelstr, _T("DEBUG"));
		break;
	default:
		_tcscpy(levelstr, _T("INFO"));
		break;
	}

#ifdef _LOG_CONSOLE
	DWORD dwWritten;
	WriteConsole(hConsole, buffer, _tcslen(buffer), &dwWritten, NULL);
#endif

	send_log_message(level, levelstr, buffer);

#ifdef _LOG_FILE
	if(fplog) {
		SYSTEMTIME st;
		GetLocalTime(&st);
		_ftprintf(fplog, _T("%04d-%02d-%02d %02d:%02d:%02d.%03d %-5s: %s\n")
			, st.wYear, st.wMonth, st.wDay, st.wHour, st.wMinute, st.wSecond, st.wMilliseconds
			, levelstr, buffer);
		fflush(fplog);
	}
#endif
}

void Logging::out_log(int level, const _TCHAR* format, ...)
{
	va_list ap;

	va_start(ap, format);
	out_logv(level, format, ap);
	va_end(ap);
}

void Logging::out_debug(const _TCHAR* format, ...)
{
#ifdef _DEBUG_LOG
	va_list ap;

	va_start(ap, format);
	out_logv(LOG_DEBUG, format, ap);
	va_end(ap);
#endif
}
