/** @file msgboard.h

	HITACHI BASIC MASTER LEVEL3 Mark5 Emulator 'EmuB-6892'
	Skelton for retropc emulator

	@author Sasaji
	@date   2012.06.26 -

	@brief [ message board ]
*/

#ifndef _MSGBOARD_H_
#define _MSGBOARD_H_

#include <windows.h>
#include <tchar.h>
#include <list>

#include <d3d9.h>
#pragma comment(lib, "d3d9.lib")
#ifdef USE_DIRECT3DX
#include <d3dx9.h>
#pragma comment(lib, "d3dx9.lib")
#endif

#include "dibsection.h"
#include "winfont.h"

class EMU;

/**
	@brief MsgBoard is the class that display various messages on the screen.
*/
class MsgBoard : public DibSection
{
private:
	HPEN   hPen;
	HBRUSH hBrush;

	COLORREF fg;
	COLORREF bg;

	SIZE szWin;

	typedef struct msg_item_st {
		_TCHAR msg[1024];
		int cnt;
	} item_t;

	typedef std::list<item_t> list_t;

	typedef struct msg_data_st {
		RECT re;	// `g
		SIZE sz;	// `悵̕A
		POINT pt;	// `̈ʒu
		int place;	// pt̊ 0: 1:E 2: 3:E
		CFont *font;
		list_t lists;
	} msg_data_t;

	msg_data_t msg;
	msg_data_t info;

	bool inited;
	bool visible;

	EMU *emu;

	// o
	void draw(HDC hdc, msg_data_t &data);
	void draw(LPDIRECT3DSURFACE9 suf, msg_data_t &data);
	// obNobt@ɕ`
	void draw_text(msg_data_t &data);
	void draw_text(HDC hdc, msg_data_t &data, int left, int top);
	// JEg_E
	void count_down(msg_data_t &data);

public:
	MsgBoard(HWND, EMU *);
	~MsgBoard();

	// 
	void InitScreen(HDC hdc, int width, int height);

	// \
	void SetVisible(bool val) {
		visible = val;
	}
	// JEg_E
	void CountDown(void);

	// bZ[Wݒ
	void Set(msg_data_t &data, const _TCHAR *str, int sec);
	// bZ[W폜
	void Delete(msg_data_t &data, const _TCHAR *str);
	// EBhETCYݒ
	void SetSize(int width, int height);

	// bZ[Wʒuݒ
	void SetMessagePos(int cx, int cy, int place);
	// bZ[Wݒ
	void SetMessage(const _TCHAR *str, int sec = 5);
	void SetMessageF(const _TCHAR *format, ...);
	// bZ[W폜
	void DeleteMessage(const _TCHAR *str);

	// ʒuݒ
	void SetInfoPos(int cx, int cy, int place);
	// ݒ
	void SetInfo(const _TCHAR *str, int sec = 2);
	void SetInfoF(const _TCHAR *format, ...);
	// 폜
	void DeleteInfo(const _TCHAR *str);

	// o
	void Draw(HDC hdc);
	void Draw(LPDIRECT3DSURFACE9 suf);

	// tHg擾
	bool SetFont();
	CFont *GetMsgFont() { return msg.font; }
	CFont *GetInfoFont() { return info.font; }
};

#endif
