/** @file keyboard.cpp

	HITACHI BASIC MASTER LEVEL3 Mark5 Emulator 'EmuB-6892'
	Skelton for retropc emulator

	@author Sasaji
	@date   2011.06.08 -

	@brief [ keyboard ]
*/

#ifdef _UNICODE
#define SI_CONVERT_GENERIC	// use generic simple.ini on all platform
#endif
#include "keyboard.h"
#include "../../SimpleIni.h"
#include "display.h"
#include "keyboard_bind.h"

#define KEYBOARD_COUNTER_CYCLE	1
#define KEYBOARD_COUNTER_MAX	(0x80 * KEYBOARD_COUNTER_CYCLE * 2)

#define KEYBIND_HEADER	"KEYBIND3"

#ifdef USE_KEY_RECORD
#define KEY_RECORD_SYSTEM_CODE	0x1000
#define KEY_RECORD_MAX			0x2000
#define KEY_RECORD_HEADER		"KEYRECORD_BML3MK5"
#endif

void KEYBOARD::initialize()
{
	key_stat = emu->key_buffer();
	kb_mode = 0x44;
	scan_code = 0;
	key_scan_code = 0;
	key_pressed = 0;
	kb_nmi = 0x7f;

	mouse_stat = emu->mouse_buffer();
	lpen_flg = 0;
	lpen_flg_prev = 0;
	lpen_bl = false;

	joy_stat = emu->joy_buffer();

	key_mod = emu->get_key_mod_ptr();
	pause_pressed = false;
	altkey_pressed = false;

	memset(vm_key_stat, 0, sizeof(vm_key_stat));
	emu->set_vm_key_buffer(vm_key_stat, KEYBIND_KEYS);

	counter = 0;
	remain_count = -1;
//	frame_counter = 0;
//	event_counter = 0;

	memcpy(scan2key_map, scan2key_defmap, sizeof(scan2key_defmap));
	memcpy(scan2joy_map, scan2joy_defmap, sizeof(scan2joy_defmap));
	for(int i=0; i<KEYBIND_PRESETS; i++) {
		memcpy(scan2key_preset_map[i], scan2key_defmap, sizeof(scan2key_defmap));
		memcpy(scan2joy_preset_map[i], scan2joy_defmap, sizeof(scan2joy_defmap));
	}

	// load keybind
	if (load_ini_file() != true) {
		load_cfg_file();
	}

	// for dialog box
	emu->set_parami(EMU::ParamVmKeyMapSize, 128);
	emu->set_paramv(EMU::ParamVmKeyMap0, (void *)kb2scan_map);
	emu->set_paramv(EMU::ParamVmKeyMap1, (void *)kb2scan_map);

	emu->set_parami(EMU::ParamVkKeyMapKeys, KEYBIND_KEYS);
	emu->set_parami(EMU::ParamVkKeyMapAssign, KEYBIND_ASSIGN);
	emu->set_paramv(EMU::ParamVkKeyDefMap0, (void *)scan2key_defmap);
	emu->set_paramv(EMU::ParamVkKeyDefMap1, (void *)scan2joy_defmap);

	emu->set_paramv(EMU::ParamVkKeyMap0, (void *)scan2key_map);
	emu->set_paramv(EMU::ParamVkKeyMap1, (void *)scan2joy_map);

	emu->set_parami(EMU::ParamVkKeyPresets, KEYBIND_PRESETS);
	for(int i=0; i<KEYBIND_PRESETS; i++) {
		emu->set_paramv(EMU::ParamVkKeyPresetMap00 + i * 2, (void *)scan2key_preset_map[i]);
		emu->set_paramv(EMU::ParamVkKeyPresetMap01 + i * 2, (void *)scan2joy_preset_map[i]);
	}

//	now_autokey = false;
#if 0
	emu->keybindbox->SetVmKeyMap(0, (uint16 *)kb2scan_map, 128);
	emu->keybindbox->SetVmKeyMap(1, (uint16 *)kb2scan_map, 128);

	emu->keybindbox->SetVkKeyDefMap(0, (uint32 *)scan2key_defmap, KEYBIND_KEYS, KEYBIND_ASSIGN);
	emu->keybindbox->SetVkKeyDefMap(1, (uint32 *)scan2joy_defmap, KEYBIND_KEYS, KEYBIND_ASSIGN);

	emu->keybindbox->SetVkKeyMap(0, (uint32 *)scan2key_map);
	emu->keybindbox->SetVkKeyMap(1, (uint32 *)scan2joy_map);

	for(int i=0; i<KEYBIND_PRESETS; i++) {
		emu->keybindbox->SetVkKeyPresetMap(0, (uint32 *)scan2key_preset_map[i], i);
		emu->keybindbox->SetVkKeyPresetMap(1, (uint32 *)scan2joy_preset_map[i], i);
	}

	emu->keybindbox->SetTitleLabel(_T("level3 key"),_T("bind"),_T("joypad"));
#endif

#ifdef USE_KEY_RECORD
	fkro = new FILEIO();
	fkri = new FILEIO();
	config.reckey_recording = false;
	config.reckey_playing = false;
	key_rec_sum_clock = 0;
	memset(vm_key_recp_stat, 0, sizeof(vm_key_recp_stat));
	memset(vm_key_recr_stat, 0, sizeof(vm_key_recr_stat));
	memset(mouse_recp_stat, 0, sizeof(mouse_recp_stat));
	memset(mouse_recr_stat, 0, sizeof(mouse_recr_stat));
#ifdef _DEBUG
	memset(vm_key_dbg_stat, 0, sizeof(vm_key_dbg_stat));
#endif
#endif

	// event
	register_frame_event(this);
}

void KEYBOARD::release()
{
#ifdef USE_KEY_RECORD
	stop_reckey();

	delete fkro;
	delete fkri;
#endif

	// save keybind
	save_keybind();
}

// ----------------------------------------------------------------------------
void KEYBOARD::save_keybind()
{
	// save keybind
	save_cfg_file();
	save_ini_file();
}

bool KEYBOARD::load_cfg_file()
{
#if 0 // cfg file is no longer supported.
	_TCHAR *app_path, file_path[_MAX_PATH];
	char header[10];

	uint16 scan_map_old16[KEYBIND_KEYS][KEYBIND_ASSIGN];

	FILEIO* fio = new FILEIO();
	app_path = emu->application_path();

	_stprintf(file_path, _T("%skeybind.cfg"), app_path);
	if(fio->Fopen(file_path, FILEIO_READ_BINARY)) {
		fio->Fread(header, sizeof(char), 8);
		if (strncmp(header, KEYBIND_HEADER, 7) == 0) {
#if 0
			fio->Fread(scan2key_map, sizeof(uint16), KEYBIND_KEYS * KEYBIND_ASSIGN);
			for(int i=0; i<KEYBIND_PRESETS; i++) {
				fio->Fread(scan2key_preset_map[i], sizeof(uint16), KEYBIND_KEYS * KEYBIND_ASSIGN);
			}
			fio->Fread(scan2joy_map, sizeof(uint16), KEYBIND_KEYS * KEYBIND_ASSIGN);
			for(int i=0; i<KEYBIND_PRESETS; i++) {
				fio->Fread(scan2joy_preset_map[i], sizeof(uint16), KEYBIND_KEYS * KEYBIND_ASSIGN);
			}
#else
			fio->Fread(scan_map_old16, sizeof(uint16), KEYBIND_KEYS * KEYBIND_ASSIGN);
			for(int kk=0; kk<KEYBIND_KEYS; kk++) {
				for(int ka=0; ka<KEYBIND_ASSIGN; ka++) {
					scan2key_map[kk][ka] = scan_map_old16[kk][ka];
				}
			}
			for(int i=0; i<KEYBIND_PRESETS; i++) {
				fio->Fread(scan_map_old16, sizeof(uint16), KEYBIND_KEYS * KEYBIND_ASSIGN);
				for(int kk=0; kk<KEYBIND_KEYS; kk++) {
					for(int ka=0; ka<KEYBIND_ASSIGN; ka++) {
						scan2key_preset_map[i][kk][ka] = scan_map_old16[kk][ka];
					}
				}
			}
			fio->Fread(scan_map_old16, sizeof(uint16), KEYBIND_KEYS * KEYBIND_ASSIGN);
			for(int kk=0; kk<KEYBIND_KEYS; kk++) {
				for(int ka=0; ka<KEYBIND_ASSIGN; ka++) {
					scan2joy_map[kk][ka] = scan_map_old16[kk][ka];
				}
			}
			for(int i=0; i<KEYBIND_PRESETS; i++) {
				fio->Fread(scan_map_old16, sizeof(uint16), KEYBIND_KEYS * KEYBIND_ASSIGN);
				for(int kk=0; kk<KEYBIND_KEYS; kk++) {
					for(int ka=0; ka<KEYBIND_ASSIGN; ka++) {
						scan2joy_preset_map[i][kk][ka] = scan_map_old16[kk][ka];
					}
				}
			}
#endif
			emu->out_log(LOG_INFO,_T("keybind.cfg was loaded."));
		} else {
			emu->out_log(LOG_WARN,_T("keybind.cfg was invalid version."));
		}
		fio->Fclose();
//	} else {
//		emu->out_log(LOG_WARN,_T("keybind.cfg couldn't be loaded."));
	}
	delete fio;

	// convert keybind map from old version to current version
	if (header[7] == '1' || header[7] == '2') {
		emu->out_log(LOG_WARN,_T("keybind.cfg is old version."));
		for (int n=0x34; n<=0x36; n+=2) {
			for (int j=0; j<2; j++) {
				scan2key_map[n][j] = scan2key_defmap[n][j];
				scan2joy_map[n][j] = scan2joy_defmap[n][j];
				for (int i=0; i<KEYBIND_PRESETS; i++) {
					scan2key_preset_map[i][n][j] = scan2key_defmap[n][j];
					scan2joy_preset_map[i][n][j] = scan2joy_defmap[n][j];
				}
			}
		}
	}
	if (header[7] == '1') {
		for (int n=0x60; n<=0x80; n++) {
			for (int j=0; j<2; j++) {
				scan2key_map[n][j] = scan2key_defmap[n][j];
				scan2joy_map[n][j] = scan2joy_defmap[n][j];
				for (int i=0; i<KEYBIND_PRESETS; i++) {
					scan2key_preset_map[i][n][j] = scan2key_defmap[n][j];
					scan2joy_preset_map[i][n][j] = scan2joy_defmap[n][j];
				}
			}
		}
	}
#endif
	return true;
}

void KEYBOARD::save_cfg_file()
{
#if 0 // cfg file is no longer supported.
	_TCHAR app_path[_MAX_PATH], file_path[_MAX_PATH];
	char header[10];

	emu->application_path(app_path);
	FILEIO* fio = new FILEIO();

	strcpy(header, KEYBIND_HEADER);
	_stprintf(file_path, _T("%skeybind.cfg"), app_path);
	if(fio->Fopen(file_path, FILEIO_WRITE_BINARY)) {
		fio->Fwrite(header, sizeof(char), 8);
		fio->Fwrite(scan2key_map, sizeof(uint16), KEYBIND_KEYS * KEYBIND_ASSIGN);
		for(int i=0; i<4; i++) {
			fio->Fwrite(scan2key_preset_map[i], sizeof(uint16), KEYBIND_KEYS * KEYBIND_ASSIGN);
		}
		fio->Fwrite(scan2joy_map, sizeof(uint16), KEYBIND_KEYS * KEYBIND_ASSIGN);
		for(int i=0; i<4; i++) {
			fio->Fwrite(scan2joy_preset_map[i], sizeof(uint16), KEYBIND_KEYS * KEYBIND_ASSIGN);
		}
		fio->Fclose();
		emu->out_log(LOG_INFO,_T("keybind.cfg was saved."));
	} else {
		emu->out_log(LOG_WARN,_T("keybind.cfg couldn't be saved."));
	}
	delete fio;
#endif
}

// ----------------------------------------------------------------------------
bool KEYBOARD::load_ini_file()
{
	const _TCHAR *app_path;
	_TCHAR file_path[_MAX_PATH];
	_TCHAR section[100];

	CSimpleIni *ini = new CSimpleIni();
#ifdef _UNICODE
	ini->SetUnicode(true);
#endif

	// load ini file
	app_path = emu->initialize_path();

	_stprintf(file_path, _T("%skeybind.ini"), app_path);
#ifdef _WIN32
	if (ini->LoadFile(file_path) == SI_OK) {
#else
	if (ini->LoadFile(UTILITY::cconv_from_native_path(file_path)) == SI_OK) {
#endif
		emu->out_log(LOG_INFO,_T("keybind.ini was loaded."));
	} else {
		emu->out_log(LOG_WARN,_T("keybind.ini couldn't be loaded."));
		delete ini;
		return false;
	}

	CSimpleIni::TNamesDepend keys;
	CSimpleIni::TNamesDepend::const_iterator ik;

	// check file version
	if (_tcscmp(ini->GetValue(_T(""), _T("Version"), _T("")), _T(VK_KEY_TYPE)) != 0) {
		emu->out_log(LOG_WARN,_T("keybind.ini may not be compatible."));
	}

	_stprintf(section, _T("Keyboard"));
	ini->GetAllKeys(section, keys);
	if (!keys.empty()) {
		memset(scan2key_map, 0, sizeof(scan2key_map));
	}
	for (ik = keys.begin(); ik != keys.end(); ik++) {
		int k = 0;
		int i = 0;
		int rc = _stscanf(ik->pItem, _T("%02x_%d"), &k, &i);
		if (rc != 2 || k < 0 || k >= KEYBIND_KEYS || i < 0 || i >= KEYBIND_ASSIGN) {
			continue;
		}
		long v = ini->GetLongValue(section, ik->pItem);
		scan2key_map[k][i] = (uint32)(v & 0xffffffff);
	}
	for (int no = 0; no < 4; no++) {
		_stprintf(section, _T("KeyboardPreset%d"), no + 1);
		ini->GetAllKeys(section, keys);
		if (!keys.empty()) {
			memset(scan2key_preset_map[no], 0, sizeof(scan2key_preset_map[no]));
		}
		for (ik = keys.begin(); ik != keys.end(); ik++) {
			int k = 0;
			int i = 0;
			int rc = _stscanf(ik->pItem, _T("%02x_%d"), &k, &i);
			if (rc != 2 || k < 0 || k >= KEYBIND_KEYS || i < 0 || i >= KEYBIND_ASSIGN) {
				continue;
			}
			long v = ini->GetLongValue(section, ik->pItem);
			scan2key_preset_map[no][k][i] = (uint32)(v & 0xffffffff);
		}
	}
	_stprintf(section, _T("Joypad"));
	ini->GetAllKeys(section, keys);
	if (!keys.empty()) {
		memset(scan2joy_map, 0, sizeof(scan2joy_map));
	}
    for (ik = keys.begin(); ik != keys.end(); ik++) {
		int k = 0;
		int i = 0;
		int rc = _stscanf(ik->pItem, _T("%02x_%d"), &k, &i);
		if (rc != 2 || k < 0 || k >= KEYBIND_KEYS || i < 0 || i >= KEYBIND_ASSIGN) {
			continue;
		}
		long v = ini->GetLongValue(section, ik->pItem);
		scan2joy_map[k][i] = (uint32)(v & 0xffffffff);
	}
	for (int no = 0; no < 4; no++) {
		_stprintf(section, _T("JoypadPreset%d"), no + 1);
		ini->GetAllKeys(section, keys);
		if (!keys.empty()) {
			memset(scan2joy_preset_map[no], 0, sizeof(scan2joy_preset_map[no]));
		}
		for (ik = keys.begin(); ik != keys.end(); ik++) {
			int k = 0;
			int i = 0;
			int rc = _stscanf(ik->pItem, _T("%02x_%d"), &k, &i);
			if (rc != 2 || k < 0 || k >= KEYBIND_KEYS || i < 0 || i >= KEYBIND_ASSIGN) {
				continue;
			}
			long v = ini->GetLongValue(section, ik->pItem);
			scan2joy_preset_map[no][k][i] = (uint32)(v & 0xffffffff);
		}
	}

	keys.clear();

	delete ini;

	return true;
}

void KEYBOARD::save_ini_file()
{
	const _TCHAR *app_path;
	_TCHAR file_path[_MAX_PATH];
	_TCHAR comment[100];
	_TCHAR section[100];
	_TCHAR key[100];

	CSimpleIni *ini = new CSimpleIni();
#ifdef _UNICODE
	ini->SetUnicode(true);
#endif

	// section
	_stprintf(comment, _T("; %s keybind file"), _T(DEVICE_NAME));
	ini->SetValue(_T(""), _T("Version"), _T(VK_KEY_TYPE), comment, true);

	_stprintf(section, _T("Keyboard"));
    for (int k = 0; k < KEYBIND_KEYS; k++) {
		for (int i = 0; i < KEYBIND_ASSIGN; i++) {
			_stprintf(key, _T("%02x_%d"), k, i);
			if (scan2key_map[k][i] != 0) {
				ini->SetLongValue(section, key, scan2key_map[k][i], NULL, true);
			}
		}
	}
    for (int no = 0; no < 4; no++) {
		_stprintf(section, _T("KeyboardPreset%d"), no + 1);
		for (int k = 0; k < KEYBIND_KEYS; k++) {
			for (int i = 0; i < KEYBIND_ASSIGN; i++) {
				_stprintf(key, _T("%02x_%d"), k, i);
				if (scan2key_preset_map[no][k][i] != 0) {
					ini->SetLongValue(section, key, scan2key_preset_map[no][k][i], NULL, true);
				}
			}
		}
	}
	_stprintf(section, _T("Joypad"));
    for (int k = 0; k < KEYBIND_KEYS; k++) {
		for (int i = 0; i < KEYBIND_ASSIGN; i++) {
			_stprintf(key, _T("%02x_%d"), k, i);
			if (scan2joy_map[k][i] != 0) {
				ini->SetLongValue(section, key, scan2joy_map[k][i], NULL, true);
			}
		}
	}
    for (int no = 0; no < 4; no++) {
		_stprintf(section, _T("JoypadPreset%d"), no + 1);
		for (int k = 0; k < KEYBIND_KEYS; k++) {
			for (int i = 0; i < KEYBIND_ASSIGN; i++) {
				_stprintf(key, _T("%02x_%d"), k, i);
				if (scan2joy_preset_map[no][k][i] != 0) {
					ini->SetLongValue(section, key, scan2joy_preset_map[no][k][i], NULL, true);
				}
			}
		}
	}

	// save ini file
	app_path = emu->initialize_path();

	_stprintf(file_path, _T("%skeybind.ini"), app_path);
#ifdef _WIN32
	if (ini->SaveFile(file_path) == SI_OK) {
#else
	if (ini->SaveFile(UTILITY::cconv_from_native_path(file_path)) == SI_OK) {
#endif
		emu->out_log(LOG_INFO,_T("keybind.ini was saved."));
	} else {
		emu->out_log(LOG_WARN,_T("keybind.ini couldn't be saved."));
	}

	delete ini;
}

// ----------------------------------------------------------------------------

void KEYBOARD::reset()
{
	kb_mode = 0;
	scan_code = 0;
	key_scan_code = 0;
	key_pressed = 0;
	kb_nmi = 0x7f;

	lpen_flg = 0;
	lpen_flg_prev = 0;
	lpen_bl = false;

	counter = 0;
	remain_count = -1;
//	frame_counter = 0;
//	event_counter = 0;
}

void KEYBOARD::write_signal(int id, uint32 data, uint32 mask)
{
	switch (id) {
		case SIG_KEYBOARD_HSYNC:
			if (data & mask) update_keyboard();
			break;
		case SIG_CPU_RESET:
			now_reset = (data & mask) ? true : false;
			reset();
			break;
	}
}

void KEYBOARD::write_io8(uint32 addr, uint32 data)
{
	if (now_reset) return;

	switch (addr & 0xffff) {
		case ADDR_LPEN_BL:
			lpen_bl = (data & 0x80) ? true : false;
//			emu->out_debug("w_lpenbl: %02x",lpen_bl);
			break;
		case ADDR_KEYBOARD:
			kb_mode = data;

			// keyborad irq interrupt
			if ((kb_mode & 0x40) && (key_scan_code & 0x80) != 0) {
				d_cpu->write_signal(SIG_CPU_IRQ, 1,1);
			}
//			emu->out_debug("w_kb: fr:%04d ct:%03x kb:%02x"
//			, frame_counter, counter, kb_mode);
			break;
	}
}

uint32 KEYBOARD::read_io8(uint32 addr)
{
	uint32 data = 0;

	if (now_reset) return data;

	switch (addr & 0xffff) {
		case ADDR_KB_NMI:
			// break key (0xffc8)
			data = kb_nmi; // & 0x80;
			kb_nmi = 0x7f;
			break;
		case ADDR_LP_FLG:
			// light pen flag (0xffcb)
			data = lpen_flg & 0x80;
			lpen_flg = 0;

//			emu->out_debug("r_lpenflg:%04d ct:%03x d:%02x"
//			, frame_counter, counter, data);
			break;
		case ADDR_KEYBOARD:
			// keyboard (0xffe0)
			data = key_scan_code;
			key_scan_code &= 0x7f;

			emu->recognized_key(data);
//			key_native_code= 0;

//			emu->out_debug("r_ks fr:%04d ct:%03x rc:%02d d:%02x k:%02x"
//			, frame_counter, counter, remain_count, data, key_scan_code);
			break;
	}
	return data;
}

void KEYBOARD::clear_scan_code(int key_counter)
{
//	if (kb_mode & 0x40) {
//		key_counter = (int)(key_counter / KEYBOARD_COUNTER_CYCLE);
//	}

	// if kb bit3 is set, enable scan_code is between 0x00 and 0x07
	scan_code = (key_counter) & ((kb_mode & 0x08) ? 0x07 : 0x7f);
	if (key_scan_code < 0x80) {
		key_scan_code = scan_code;
	}
}

void KEYBOARD::update_scan_code(int key_counter)
{
	bool pressed = false;

	if (scan_code == 0x7f) {
		if (key_pressed == 1) {
			pressed = true;
			scan_code |= 0x80;
			remain_count = KEYBOARD_COUNTER_MAX * 8;
		}
		if (key_pressed > 0) key_pressed--;
	} else if (scan_code >= 0x60) {
		// no assaign
	} else {
		pressed = pressing_key(scan_code);
		if (pressed == true) {
			scan_code |= 0x80;
			key_pressed = 2;
			remain_count = KEYBOARD_COUNTER_MAX * 8;
		}
	}

	if (key_scan_code < 0x80) {
		key_scan_code = scan_code;
	} else if (scan_code >= 0x80) {
		key_scan_code = scan_code;
	}

	// keyborad irq interrupt
	if ((kb_mode & 0x40) && (pressed == true)) {
		d_cpu->write_signal(SIG_CPU_IRQ, 1,1);
	}

//	pressed = false;

//	emu->out_debug("kbd fr:%04d ct:%03x rc:%02d s:%02x k:%02x"
//	, frame_counter, counter, remain_count, scan_code, key_scan_code);
}

void KEYBOARD::update_special_key()
{
	bool pressed = false;

	for(int code = 0x7f; code <= 0x80; code++) {
		pressed = pressing_key(code);
		switch(code) {
			case 0x7f:
				// reset switch
				if (pressed != now_reset) {
					d_board->write_signal(SIG_CPU_RESET, pressed ? 1 : 0, 1);
				}
				break;
			case 0x80:
				// break key
				if (pressed == true) {
					if (!(kb_nmi & 0x80)) {
						kb_nmi = 0xff;
						// keyborad nmi interrupt
						if ((kb_mode & 0x80)) {
							d_cpu->write_signal(SIG_CPU_NMI, 1,1);
						}
					}
				} else {
					kb_nmi = 0x7f;
				}
				break;
		}
	}
}

void KEYBOARD::update_light_pen()
{
	if (lpen_bl && (config.show_dlg & 0x10)) {
		recording_mouse_status();

		lpen_flg = (mouse_stat[2] & 1) ? 0x80 : 0;
		if (lpen_flg != 0) {
			// push light pen
			// LPSTB signal
			d_disp->write_signal(SIG_DISPLAY_LPSTB
				,((mouse_stat[0] & 0xffff) << 16) | (mouse_stat[1] & 0xffff)
				, 0xffffffff);

			// IRQ interrupt
			d_cpu->write_signal(SIG_CPU_IRQ, 1,1);

		}
#ifdef USE_KEY_RECORD
		if (config.reckey_playing) {
			lpen_flg = (mouse_recp_stat[2] & 1) ? 0x80 : 0;
			if (lpen_flg != 0) {
				// push light pen
				// LPSTB signal
				d_disp->write_signal(SIG_DISPLAY_LPSTB
					,((mouse_recp_stat[0] & 0xffff) << 16) | (mouse_recp_stat[1] & 0xffff)
					, 0xffffffff);

				// IRQ interrupt
				d_cpu->write_signal(SIG_CPU_IRQ, 1,1);

			}
		}
#endif
		lpen_flg_prev = lpen_flg;

	}
//	emu->out_debug("lpen bl:%02x flg:%02x x:%03d y:%03d reg:%02x%02x",lpen_bl,lpen_flg, mouse_stat[0], mouse_stat[1], regs[16], regs[17]);
}

void KEYBOARD::update_system_key()
{
	bool pressed = false;

	for(int code = 0x60; code <= 0x61; code++) {
		pressed = pressing_key(code);
		switch(code) {
			case 0x60:
				// pause
				if (pressed && pause_pressed == false) {
					emu->set_pause(2, emu->get_pause(2) ? false : true);
				}
				pause_pressed = pressed;
				break;
			case 0x61:
				// alt
				altkey_pressed = pressed;
				*key_mod = (altkey_pressed) ? (*key_mod) | KEY_MOD_ALT_KEY : (*key_mod) & ~KEY_MOD_ALT_KEY;
				break;
		}
	}
}

// ----------------------------------------------------------------------------

bool KEYBOARD::pressing_key(int key_code)
{
	uint32 code = 0;
	bool pressed = false;
	int i = 0;

	for(i=0; i<2; i++) {
		code = scan2key_map[key_code][i];
		if (code == 0) break;

		// key pressed ?
		if (key_stat[code]) {
			pressed = true;
			break;
		}
	}
	if (pressed != true) {
		for(i=0; i<2; i++) {
			code = scan2joy_map[key_code][i];
			if (code == 0) continue;

			// joypad pressed ?
			// allow key or button
			if (
				((scan2joy_map[0x81][0] & 1) == 0 && ((joy_stat[i] & 0xf & code) == code || (joy_stat[i] & 0xfffffff0 & code) == code))
			||  ((scan2joy_map[0x81][0] & 1) != 0 && ((joy_stat[i] & 0xf) == code || (joy_stat[i] & 0xfffffff0) == code))
			) {
				pressed = true;
				break;
			}
		}
	}
	if (pressed != true) {
		// auto key pressed ?
		if (vm_key_stat[key_code]) {
			pressed = true;
		}
	}
#ifdef USE_KEY_RECORD
	if (pressed != true) {
		// record key pressed ?
		if (vm_key_recp_stat[key_code]) {
			pressed = true;
		}
	}
#ifdef _DEBUG
	if ((vm_key_dbg_stat[key_code] != 0) != pressed) {
#ifndef __MINGW32__
		emu->out_log(LOG_DEBUG, _T("   Key0 %02x  c:%llu kc:%d kr:%d %s"), key_code, current_clock(), counter, remain_count, pressed ? _T("ON") : _T("OFF"));
#else
		emu->out_log(LOG_DEBUG, _T("   Key0 %02x  c:%I64u kc:%d kr:%d %s"), key_code, current_clock(), counter, remain_count, pressed ? _T("ON") : _T("OFF"));
#endif
		vm_key_dbg_stat[key_code] = (pressed ? 1 : 0);
	}
#endif
	recording_keys(key_code, pressed);
#endif

	return pressed;
}

// ----------------------------------------------------------------------------
// event handler
// ----------------------------------------------------------------------------

void KEYBOARD::event_frame()
{
#ifdef USE_KEY_RECORD
	// read record key if need
	reading_keys(1);
#endif

	// reset switch & break key
	update_special_key();
	// other system keys
	update_system_key();
	// Cgy
	update_light_pen();
}

void KEYBOARD::event_callback(int event_id, int err)
{
}

// ----------------------------------------------------------------------------
void KEYBOARD::update_keyboard()
{
//	if (event_counter == 0) {
//		frame_counter = (frame_counter + 1) % 10000;
//	}
//	event_counter++;
//	if (event_counter >= KEYBOARD_COUNTER_MAX) event_counter = 0;

	if (remain_count == 0) {
		key_scan_code &= 0x7f;
		counter = (counter - 2) % KEYBOARD_COUNTER_MAX;
	}
	if (remain_count >= 0) remain_count--;

#ifdef USE_KEY_RECORD
	// read record key if need
//	if ((counter & 1) != 0) {
		reading_keys(0);
//	}
#endif
//	if ((counter % 16) == 1) emu->out_debug(_T("KEY COUNTER: %lld %03d"),current_clock(), counter);
	// L[ĂȂƂɃL[XL
	if ((key_scan_code & 0x80) == 0) {
		if ((counter & 1) == 0) {
			clear_scan_code(counter >> 1);
		} else {
			update_scan_code(counter >> 1);
		}
		if (!now_reset) counter++;
		if (counter >= KEYBOARD_COUNTER_MAX) counter = 0;
	}
}

// ----------------------------------------------------------------------------
void KEYBOARD::system_key_down(int code)
{
	recording_system_keys(code, 1);
}

void KEYBOARD::system_key_up(int code)
{
	recording_system_keys(code, 0);
}

// ----------------------------------------------------------------------------
void KEYBOARD::reading_keys(int num)
{
#ifdef USE_KEY_RECORD
	if (config.reckey_playing) {
		uint64 now_clk = current_clock();
		int cols = -1;
		uint64 clk = 0;
		int type = -1;
		int code[3];
		char *p = NULL;

		while(config.reckey_playing) {
			if (0x30 <= rec_key_rec_buff[0] && rec_key_rec_buff[0] <= 0x39) {
#ifndef __MINGW32__
				cols = sscanf(rec_key_rec_buff,"%llu:%d:",&clk,&type);
#else
				cols = sscanf(rec_key_rec_buff,"%I64u:%d:",&clk,&type);
#endif
				if (clk + key_rec_sum_clock < 0) {
					clk = 0;
				} else {
					clk += key_rec_sum_clock;
				}
//				emu->out_log(LOG_DEBUG,_T("RecKey:n:%llu:%llu"),now_clk,clk);
				if ((now_clk + 1) < clk) {
					break;
				}
//				emu->out_log(LOG_DEBUG,_T("RecKey:keyin"));

				if (cols == 2 && clk > 0) {
					p = strchr(rec_key_rec_buff, ':');
					p = strchr(p+1, ':');
					switch(type) {
					case 1:
						// key
						cols = sscanf(p,":%x:%d",&code[0],&code[1]);
						if (cols == 2 && code[0] >= 0) {
#ifdef _DEBUG
#ifndef __MINGW32__
							emu->out_log(LOG_DEBUG, _T("RecKey%d %02x nc:%llu %c %llu nk:%d %c %d %s"), num, code[0]
#else
							emu->out_log(LOG_DEBUG, _T("RecKey%d %02x nc:%I64u %c %I64u nk:%d %c %d %s"), num, code[0]
#endif
								, now_clk, (now_clk == clk ? _T('=') : _T('!')), clk
								, counter, ((code[0] << 1) + 1 == counter ? _T('=') : _T('!')), (code[0] << 1) + 1
								, (code[1] & 1) ? _T("ON") : _T("OFF"));
#endif
							if (code[0] < KEYBIND_KEYS) {
								// normal key
								vm_key_recp_stat[code[0]]=(code[1] & 1);
								emu->press_global_key(false);
								if (num == 0 && (clk < now_clk || ((code[0] << 1) + 1) < counter)) {
									// adjust timing
									counter = (((code[0] << 1) & ~1) | (counter & 1));
//									remain_count = code[2];
									scan_code = ((counter >> 1) & ((kb_mode & 0x08) ? 0x07 : 0x7f));

									emu->out_debug(_T("RecKey%d %02x adjust k:%d s:%x"), num, code[0], counter, scan_code);
								}
							} else if (code[0] >= KEY_RECORD_SYSTEM_CODE && code[0] < KEY_RECORD_MAX) {
								if (code[1] & 1) {
									// global key
									emu->system_key_down(code[0] & 0xfff);
									emu->execute_global_keys(code[0] & 0xfff, 0);
								}
							}
						}
						break;
					case 2:
						// mouse
						cols = sscanf(p,":%d:%d:%x",&code[0],&code[1],&code[2]);
						if (cols == 3) {
							memcpy(mouse_recp_stat, code, sizeof(mouse_recp_stat));
						}
						break;
					}
				}
			}
			// read next data
			if (fkri->Fgets(rec_key_rec_buff, sizeof(rec_key_rec_buff)) == NULL) {
				stop_reckey(true, false);
				break;
			}
		}
	}
#endif
}

void KEYBOARD::recording_keys(int code, bool pressed)
{
#ifdef USE_KEY_RECORD
	if (config.reckey_recording) {
		if (pressed && vm_key_recr_stat[code] == 0) {
#ifndef __MINGW32__
			sprintf(rec_key_tmp_buff,"%llu:1:%04x:1\n", current_clock(), code);
#else
			sprintf(rec_key_tmp_buff,"%I64u:1:%04x:1\n", current_clock(), code);
#endif
			fkro->Fwrite(rec_key_tmp_buff, strlen(rec_key_tmp_buff), 1);
			vm_key_recr_stat[code] = 1;
		} else if (!pressed && vm_key_recr_stat[code] != 0) {
#ifndef __MINGW32__
			sprintf(rec_key_tmp_buff,"%llu:1:%04x:0\n", current_clock(), code);
#else
			sprintf(rec_key_tmp_buff,"%I64u:1:%04x:0\n", current_clock(), code);
#endif
			fkro->Fwrite(rec_key_tmp_buff, strlen(rec_key_tmp_buff), 1);
			vm_key_recr_stat[code] = 0;
		}
	}
#endif
}

void KEYBOARD::recording_system_keys(int code, int pressed)
{
#ifdef USE_KEY_RECORD
	if (config.reckey_recording) {
		code |= KEY_RECORD_SYSTEM_CODE;
		if (code < 0x2000) {
#ifndef __MINGW32__
			sprintf(rec_key_tmp_buff,"%llu:1:%04x:%d\n", current_clock(), code, pressed);
#else
			sprintf(rec_key_tmp_buff,"%I64u:1:%04x:%d\n", current_clock(), code, pressed);
#endif
			fkro->Fwrite(rec_key_tmp_buff, strlen(rec_key_tmp_buff), 1);
		}
	}
#endif
}

void KEYBOARD::recording_mouse_status()
{
#ifdef USE_KEY_RECORD
	if (config.reckey_recording) {
		if (mouse_stat[2] != mouse_recr_stat[2]) {
#ifndef __MINGW32__
			sprintf(rec_key_tmp_buff,"%llu:2:%d:%d:%x\n", current_clock(), mouse_stat[0], mouse_stat[1], mouse_stat[2] & 3);
#else
			sprintf(rec_key_tmp_buff,"%I64u:2:%d:%d:%x\n", current_clock(), mouse_stat[0], mouse_stat[1], mouse_stat[2] & 3);
#endif
			fkro->Fwrite(rec_key_tmp_buff, strlen(rec_key_tmp_buff), 1);
			memcpy(mouse_recr_stat, mouse_stat, sizeof(mouse_recr_stat));
		}
	}
#endif
}

// ----------------------------------------------------------------------------
#ifdef USE_KEY_RECORD
bool KEYBOARD::play_reckey(_TCHAR* filename)
{
	stop_reckey(true, false);

	char *p = NULL;
	int cols;
	int version;
	uint64 start_clock;

	fkri->Fopen(filename ,FILEIO_READ_ASCII);
	if (fkri->IsOpened()) {
		// check header
		p = fkri->Fgets(rec_key_tmp_buff, sizeof(rec_key_tmp_buff));
		if (p == NULL || strncmp(rec_key_tmp_buff, KEY_RECORD_HEADER, strlen(KEY_RECORD_HEADER)) != 0) {
			// error
			emu->out_log(LOG_ERROR, _T("This is not record key file."));
			goto FIN;
		}
		p = fkri->Fgets(rec_key_tmp_buff, sizeof(rec_key_tmp_buff));
		cols = sscanf(rec_key_tmp_buff, "Version:%d", &version);
		if (p == NULL || cols != 1 || version != 1) {
			// error
			emu->out_log(LOG_ERROR, _T("Record key file is invalid version."));
			goto FIN;
		}
		p = fkri->Fgets(rec_key_tmp_buff, sizeof(rec_key_tmp_buff));
#ifndef __MINGW32__
		cols = sscanf(rec_key_tmp_buff, "StartClock:%llu", &start_clock);
#else
		cols = sscanf(rec_key_tmp_buff, "StartClock:%I64u", &start_clock);
#endif
		if (p == NULL || cols != 1) {
			// error
			emu->out_log(LOG_ERROR, _T("Record key file has invalid parameter."));
			goto FIN;
		}
//		emu->out_log(LOG_DEBUG, _T("RecordKey StartClock:%llu"), start_clock);

		_TCHAR base_path[_MAX_PATH];

		UTILITY::get_dir_and_basename(filename, base_path, NULL);
		emu->out_debug(_T("BasePath:%s"),base_path);

		// parse optional parameters
		_TCHAR *state_file = NULL;
		_TCHAR *tape_file = NULL;
		_TCHAR *disk_file[MAX_DRIVE];
		int disk_bank_num[MAX_DRIVE];
		bool tape_playing = true;
		int drv = 0;

		for(drv = 0; drv < MAX_DRIVE; drv++) {
			disk_file[drv] = NULL;
		}

		for(;;) {
			p = fkri->Fgets(rec_key_tmp_buff, sizeof(rec_key_tmp_buff));
			if (p == NULL || p[0] == '\r' || p[0] == '\n') {
				break;
			}
			if (strncmp(rec_key_tmp_buff, "StateFile:", 10) == 0) {
				if (!state_file) state_file = new _TCHAR[_MAX_PATH];
				memset(state_file, 0, sizeof(_TCHAR) * _MAX_PATH);
				_tcsncpy(state_file, UTILITY::conv_to_native_path(&rec_key_tmp_buff[10], _MAX_PATH), _MAX_PATH-1);
				UTILITY::convert_path_separator(state_file);
				UTILITY::make_absolute_path(base_path, state_file);
				emu->out_debug(_T("StateFile:%s"),state_file);
			} else if (strncmp(rec_key_tmp_buff, "TapeFile:", 9) == 0) {
				if (!tape_file) tape_file = new _TCHAR[_MAX_PATH];
				memset(tape_file, 0, sizeof(_TCHAR) * _MAX_PATH);
				_tcsncpy(tape_file, UTILITY::conv_to_native_path(&rec_key_tmp_buff[9], _MAX_PATH), _MAX_PATH-1);
				UTILITY::convert_path_separator(tape_file);
				UTILITY::make_absolute_path(base_path, tape_file);
				emu->out_debug(_T("TapeFile:%s"),tape_file);
			} else if (strncmp(rec_key_tmp_buff, "TapeType:", 9) == 0) {
				if (strncmp(&rec_key_tmp_buff[9], "Rec", 3) == 0) {
					tape_playing = false;
				}
			} else if (sscanf(rec_key_tmp_buff, "Disk%dFile:", &drv) == 1) {
				if (0 <= drv && drv < MAX_DRIVE) {
					if (!disk_file[drv]) disk_file[drv] = new _TCHAR[_MAX_PATH];
					memset(disk_file[drv], 0, sizeof(_TCHAR) * _MAX_PATH);
					disk_bank_num[drv] = 0;
					_tcsncpy(disk_file[drv], UTILITY::conv_to_native_path(strchr(rec_key_tmp_buff, ':')+1, _MAX_PATH), _MAX_PATH-1);
					get_number_in_path(disk_file[drv], &disk_bank_num[drv]);
					UTILITY::convert_path_separator(disk_file[drv]);
					UTILITY::make_absolute_path(base_path, disk_file[drv]);
					emu->out_debug(_T("Disk%dFile:%s"),drv,disk_file[drv]);
				}
			}
		}

		// open files
		if (state_file) {
			emu->load_state(state_file);
		}
		if (tape_file) {
			if (tape_playing) {
				emu->play_datarec(tape_file);
			} else {
				emu->rec_datarec(tape_file);
			}
		}
		for(drv = 0; drv < MAX_DRIVE; drv++) {
			if (disk_file[drv]) {
				emu->open_disk_by_bank_num(drv, disk_file[drv], disk_bank_num[drv], 0, false);
			}
		}

		delete [] state_file;
		delete [] tape_file;
		for(drv = 0; drv < MAX_DRIVE; drv++) {
			delete [] disk_file[drv];
		}

		// adjust start clock
		key_rec_sum_clock = current_clock() - start_clock;
		// check ok
		config.reckey_playing = true;

		// read first data
		memset(rec_key_rec_buff, 0, sizeof(rec_key_rec_buff));
		fkri->Fgets(rec_key_rec_buff, sizeof(rec_key_rec_buff));
#ifndef __MINGW32__
		emu->out_debug(_T("RecKeyStart: c:%llu s:%llu"), current_clock(), start_clock);
#else
		emu->out_debug(_T("RecKeyStart: c:%I64u s:%I64u"), current_clock(), start_clock);
#endif
	}
FIN:
	return config.reckey_playing;
}

bool KEYBOARD::record_reckey(_TCHAR* filename)
{
	stop_reckey(false, true);

	fkro->Fopen(filename ,FILEIO_WRITE_ASCII);
	if (fkro->IsOpened()) {
		config.reckey_recording = true;

		_TCHAR base_path[_MAX_PATH];
		_TCHAR path[_MAX_PATH];

		UTILITY::get_dir_and_basename(filename, base_path, NULL);

		// write header
		sprintf(rec_key_tmp_buff, "%s\n", KEY_RECORD_HEADER);
		fkro->Fwrite(rec_key_tmp_buff, strlen(rec_key_tmp_buff), 1);
		sprintf(rec_key_tmp_buff, "Version:%d\n", 1);
		fkro->Fwrite(rec_key_tmp_buff, strlen(rec_key_tmp_buff), 1);
#ifndef __MINGW32__
		sprintf(rec_key_tmp_buff, "StartClock:%llu\n", current_clock());
#else
		sprintf(rec_key_tmp_buff, "StartClock:%I64u\n", current_clock());
#endif
		fkro->Fwrite(rec_key_tmp_buff, strlen(rec_key_tmp_buff), 1);

		if (_tcslen(config.saved_state_path) > 0) {
			_tcscpy(path, config.saved_state_path);
			UTILITY::make_relative_path(base_path, path);
			sprintf(rec_key_tmp_buff, "StateFile:%s\n", UTILITY::cconv_from_native_path(path, _MAX_PATH));
			fkro->Fwrite(rec_key_tmp_buff, strlen(rec_key_tmp_buff), 1);
		}

		if (_tcslen(config.opened_datarec_path.path) > 0) {
			_tcscpy(path, config.opened_datarec_path.path);
			UTILITY::make_relative_path(base_path, path);
			sprintf(rec_key_tmp_buff, "TapeFile:%s\n", UTILITY::cconv_from_native_path(path, _MAX_PATH));
			fkro->Fwrite(rec_key_tmp_buff, strlen(rec_key_tmp_buff), 1);
			if (emu->datarec_opened(true)) {
				sprintf(rec_key_tmp_buff, "TapeType:Play\n");
				fkro->Fwrite(rec_key_tmp_buff, strlen(rec_key_tmp_buff), 1);
			} else if (emu->datarec_opened(false)) {
				sprintf(rec_key_tmp_buff, "TapeType:Rec\n");
				fkro->Fwrite(rec_key_tmp_buff, strlen(rec_key_tmp_buff), 1);
			}
		}

		for(int drv = 0; drv < MAX_DRIVE; drv++) {
			if (_tcslen(config.opened_disk_path[drv].path) > 0) {
				_tcscpy(path, config.opened_disk_path[drv].path);
				UTILITY::make_relative_path(base_path, path);
				set_number_in_path(path, config.opened_disk_path[drv].num);
				sprintf(rec_key_tmp_buff, "Disk%dFile:%s\n", drv, UTILITY::cconv_from_native_path(path, _MAX_PATH));
				fkro->Fwrite(rec_key_tmp_buff, strlen(rec_key_tmp_buff), 1);
			}
		}

		sprintf(rec_key_tmp_buff, "\n");
		fkro->Fwrite(rec_key_tmp_buff, strlen(rec_key_tmp_buff), 1);
	}
	return config.reckey_recording;
}

void KEYBOARD::stop_reckey(bool stop_play, bool stop_record)
{
	if (stop_play && fkri) {
		fkri->Fclose();
		config.reckey_playing = false;
		memset(vm_key_recp_stat, 0, sizeof(vm_key_recp_stat));
		memset(mouse_recp_stat, 0, sizeof(mouse_recp_stat));
	}
	if (stop_record && fkro) {
		fkro->Fclose();
		config.reckey_recording = false;
		memset(vm_key_recr_stat, 0, sizeof(vm_key_recr_stat));
		memset(mouse_recr_stat, 0, sizeof(mouse_recr_stat));
	}
}
#endif

// ----------------------------------------------------------------------------
void KEYBOARD::update_config()
{
//	now_autokey = emu->now_auto_key();
}

// ----------------------------------------------------------------------------
void KEYBOARD::save_state(FILEIO *fio)
{
	struct vm_state_st vm_state;

	//
	vm_state_ident.version = 2;
	vm_state_ident.size = sizeof(vm_state_ident) + sizeof(vm_state);

	// copy values
	memset(&vm_state, 0, sizeof(vm_state));
	vm_state.kb_mode = kb_mode;
	vm_state.kb_nmi = kb_nmi;
	vm_state.lpen_flg = lpen_flg;
	vm_state.lpen_flg_prev = lpen_flg_prev;
	vm_state.lpen_bl = lpen_bl ? 1 : 0;
	vm_state.lpen_bl |= (config.show_dlg & 0x10) ? 2 : 0;	// add version 2
	vm_state.counter = counter;
	vm_state.remain_count = remain_count;

	fio->Fwrite(&vm_state_ident, sizeof(vm_state_ident), 1);
	fio->Fwrite(&vm_state, sizeof(vm_state), 1);
}

void KEYBOARD::load_state(FILEIO *fio)
{
	vm_state_ident_t vm_state_i;
	struct vm_state_st vm_state;

	READ_STATE_CHUNK(fio, vm_state_i, vm_state);

	kb_mode = vm_state.kb_mode;
	kb_nmi = vm_state.kb_nmi;
	lpen_flg = vm_state.lpen_flg;
	lpen_flg_prev = vm_state.lpen_flg_prev;
	lpen_bl = (vm_state.lpen_bl & 1) ? true : false;
	counter = vm_state.counter;
	remain_count = vm_state.remain_count;
	if (vm_state_i.version >= 2) {
		config.show_dlg = (vm_state.lpen_bl & 2) ? (config.show_dlg | 0x10) : (config.show_dlg & ~0x10);
	}
}
