/// @file parsewav_buffer.h
///
/// @brief ϊf[^ێpXg[obt@
///

#ifndef _PARSEWAV_BUFFER_H_
#define _PARSEWAV_BUFFER_H_

#include <stdio.h>

#define TERMINATOR_SIZE		4
#define DATA_BUFFER_SIZE	(128 + TERMINATOR_SIZE)

/// ϊf[^ێpXg[obt@
template <class TYPE>
class Buf_t {
public:
	TYPE   *data;
	uint32 *ms;
	uint8  *attr;
	size_t  size;
	size_t  start_pos;
	size_t  w_pos;
	size_t  r_pos;
	size_t  total_w_pos;
	size_t  total_r_pos;
	bool    last;

	Buf_t(bool use_ms, bool use_attr, size_t init_size) {
		size = init_size - TERMINATOR_SIZE;
		ms = NULL;
		attr = NULL;
		data = new TYPE[init_size];
		if (use_ms) ms = new uint32[init_size];
		if (use_attr) attr = new uint8[init_size];
	}
	Buf_t(bool use_ms, bool use_attr) {
		size = DATA_BUFFER_SIZE - TERMINATOR_SIZE;
		ms = NULL;
		attr = NULL;
		data = new TYPE[DATA_BUFFER_SIZE];
		if (use_ms) ms = new uint32[DATA_BUFFER_SIZE];
		if (use_attr) attr = new uint8[DATA_BUFFER_SIZE];
	}
	Buf_t(bool use_ms) {
		size = DATA_BUFFER_SIZE - TERMINATOR_SIZE;
		ms = NULL;
		attr = NULL;
		data = new TYPE[DATA_BUFFER_SIZE];
		if (use_ms) ms = new uint32[DATA_BUFFER_SIZE];
	}
	Buf_t() {
		size = DATA_BUFFER_SIZE - TERMINATOR_SIZE;
		ms = NULL;
		attr = NULL;
		data = new TYPE[DATA_BUFFER_SIZE];
	}
	~Buf_t() {
		if (attr) delete [] attr;
		if (ms) delete [] ms;
		delete [] data;
	}
	void init(void) {
		clear();
		total_w_pos = 0;
		total_r_pos = 0;
		last = false;
	}
	void clear(void) {
		for(size_t i=0; i<(size + TERMINATOR_SIZE); i++) {
			data[i] = 0;
			if (ms) ms[i] = 0;
			if (attr) attr[i] = 0;
		}
		start_pos = 0;
		w_pos = 0;
		r_pos = 0;
	}
	// obt@Ƀf[^ǉ
	void add(TYPE c) {
		add(c, 0, 0);
	}
	void add(TYPE c, uint32 ss) {
		add(c, ss, 0);
	}
	void add(TYPE c, uint32 ss, uint8 att) {
		if (w_pos >= size) return;
		data[w_pos] = c;
		if (ms) ms[w_pos] = ss;
		if (attr) attr[w_pos] = att;
		w_pos++;
		total_w_pos++;
	}
	// obt@Ƀf[^ǉ
	int adds(TYPE *dat) {
		return adds(dat, 0, 0);
	}
	int adds(TYPE *dat, uint32 ss) {
		return adds(dat, ss, 0);
	}
	int adds(TYPE *dat, uint32 ss, uint8 att) {
		int len = 0;
		while(*dat != 0 && w_pos<size) {
			add(*dat, ss, att);
			dat++;
			len++;
		}
		return len;
	}
	// obt@Ƀf[^ǉ
	int addns(TYPE *dat, size_t dat_len) {
		return addns(dat, 0, 0, dat_len);
	}
	int addns(TYPE *dat, uint32 ss, size_t dat_len) {
		return addns(dat, ss, 0, dat_len);
	}
	int addns(TYPE *dat, uint32 ss, uint8 att, size_t dat_len) {
		int len = 0;
		for(size_t i=0; i<dat_len && w_pos<size; i++) {
			add(dat[i], ss, att);
			len++;
		}
		return len;
	}
	// obt@ɓf[^ǉ
	int set(TYPE c, uint32 ss, size_t dat_len) {
		int len = 0;
		for(size_t i=0; i<dat_len && w_pos<size; i++) {
			add(c, ss);
			len++;
		}
		return len;
	}
	// f[^𓾂
	uint32 get_ms(int offset) {
		if (ms) {
			return ms[r_pos + offset];
		} else {
			return 0;
		}
	}
	uint8 get_attr(int offset) {
		if (attr) {
			return attr[r_pos + offset];
		} else {
			return 0;
		}
	}
	TYPE get(int offset) {
		return data[r_pos + offset];
	}
	void get(TYPE &c, uint32 &ss) {
		if (r_pos >= w_pos) {
			c = get(-1);
			ss = get_ms(-1);
		} else {
			c = get(0);
			ss = get_ms(0);
			r_pos++;
			total_r_pos++;
		}
	}
	// ̃f[^obt@̐擪ɃVtgB
	void shift() {
		if (r_pos == 0) return;
		for(int i=0; i<(int)(w_pos - r_pos); i++) {
			data[i] = data[i+r_pos];
			if (ms) ms[i] = ms[i+r_pos];
			if (attr) attr[i] = attr[i+r_pos];
		}
		w_pos = w_pos - r_pos;
		for(size_t i=w_pos; i<(size + TERMINATOR_SIZE); i++) {
			data[i] = 0;
			if (ms) ms[i] = 0;
			if (attr) attr[i] = 0;
		}
		r_pos = 0;
		start_pos = w_pos;
	}
	TYPE *startptr(void) {
		return &data[start_pos];
	}
	int length(void) {
		return w_pos - start_pos;
	}
	TYPE *curptr(int offset) {
		return &data[r_pos+offset];
	}
	int sublen(void) {
		return w_pos - r_pos;
	}
	bool isFull(int offset) {
		return (w_pos + offset >= size) ? true : false;
	}
	bool isEOB(int offset) {
		return (r_pos + offset >= w_pos) ? true : false;
	}
	// r
	int cmp(int offset, TYPE *dat, size_t len) {
		return memcmp((void *)&data[r_pos + offset], (void *)dat, len * sizeof(TYPE));
	}
	// T(char̂)
	int find(int offset, char *str) {
		int idx = -1;
		char *p = strstr(&data[r_pos + offset], str);
		if (p != NULL) {
			idx = (int)(p - &data[r_pos + offset]);
		}
		return idx;
	}
};

#endif
