/** @file win32_input_keysym.cpp

	Skelton for retropc emulator

	@author Sasaji
	@date   2014.04.01

	@brief [ win32 input keysym ]
*/

#include "emu.h"
#include "vm/vm.h"

/// notify accel key
int EMU::system_key_down(int code)
{
	if (vm) {
		code = translate_global_key(code);
		vm->system_key_down(code);
	}
	return code;
}
void EMU::system_key_up(int code)
{
	if (vm) {
		code = translate_global_key(code);
		vm->system_key_up(code);
	}
}

/// execute accel key
void EMU::execute_global_keys(int code, long status)
{
	execute_global_keys(main_window_handle, code, status);
}

void EMU::execute_global_keys(HWND hWnd, int code, long status)
{
	code = translate_global_key(code);
	switch(code) {
		case VK_RETURN:
			SendMessage(hWnd, WM_COMMAND, MAKELONG(ID_ACCEL_SCREEN, 1), 0);
			break;
		case '0':
			PostMessage(hWnd, WM_COMMAND, MAKELONG(ID_SYNC_IRQ, 1), 0);
			break;
		case '1':
			PostMessage(hWnd, WM_COMMAND, MAKELONG(ID_CPU_POWER1, 1), 0);
			break;
		case '2':
			PostMessage(hWnd, WM_COMMAND, MAKELONG(ID_CPU_POWER2, 1), 0);
			break;
		case '3':
			PostMessage(hWnd, WM_COMMAND, MAKELONG(ID_CPU_POWER3, 1), 0);
			break;
		case '4':
			PostMessage(hWnd, WM_COMMAND, MAKELONG(ID_CPU_POWER4, 1), 0);
			break;
		case '5':
			PostMessage(hWnd, WM_COMMAND, MAKELONG(ID_CPU_POWER5, 1), 0);
			break;
		case '9':
			PostMessage(hWnd, WM_COMMAND, MAKELONG(ID_CPU_POWER0, 1), 0);
			break;
		case 'C':
			PostMessage(hWnd, WM_COMMAND, MAKELONG(ID_OPTIONS_CONFIG, 1), 0);
			break;
		case 'E':
			PostMessage(hWnd, WM_COMMAND, MAKELONG(ID_RECKEY_PLAY, 1), 0);
			break;
		case 'F':
			SendMessage(hWnd, WM_COMMAND, MAKELONG(ID_OPTIONS_FDD_TYPE_A, 1), 0);
			break;
		case 'K':
			PostMessage(hWnd, WM_COMMAND, MAKELONG(ID_OPTIONS_KEYBIND, 1), 0);
			break;
		case 'L':
			SendMessage(hWnd, WM_COMMAND, MAKELONG(ID_OPTIONS_LEDBOX_SHOW, 1), 0);
			break;
		case 'M':
			SendMessage(hWnd, WM_COMMAND, MAKELONG(ID_DIPSWITCH3, 1), 0);
			break;
		case 'O':
			PostMessage(hWnd, WM_COMMAND, MAKELONG(ID_LOAD_STATE, 1), 0);
			break;
		case 'P':
			SendMessage(hWnd, WM_COMMAND, MAKELONG(ID_OPTIONS_LEDBOX_POS_A, 1), 0);
			break;
		case 'Q':
			PostMessage(hWnd, WM_COMMAND, MAKELONG(ID_PAUSE, 1), 0);
			break;
		case 'R':
			SendMessage(hWnd, WM_COMMAND, MAKELONG(ID_WARM_RESET, 1), 0);
			break;
		case 'S':
			SendMessage(hWnd, WM_COMMAND, MAKELONG(ID_SCREEN_SCANLINE, 1), 0);
			break;
		case 'T':
			SendMessage(hWnd, WM_COMMAND, MAKELONG(ID_SCREEN_AFTERIMAGE_A, 1), 0);
			break;
		case 'V':
			PostMessage(hWnd, WM_COMMAND, MAKELONG(ID_SOUND_VOLUME, 1), 0);
			break;
		case 'Y':
			SendMessage(hWnd, WM_COMMAND, MAKELONG(ID_SCREEN_D3D_A, 1), 0);
			break;
		case 'Z':
			SendMessage(hWnd, WM_COMMAND, MAKELONG(ID_OPTIONS_MSGBOARD, 1), 0);
			break;
		case 0x11c:	//	translated from VK_F3
			SendMessage(hWnd, WM_COMMAND, MAKELONG(ID_RESET, 1), 0);
			break;
		case 0x11d:	//  translated from VK_F4
			SendMessage(hWnd, WM_COMMAND, MAKELONG(ID_EXIT, 1), 0);
			break;
		case 0x11e:	//  translated from VK_F5
			SendMessage(hWnd, WM_COMMAND, MAKELONG(ID_REWIND_DATAREC, 1), 0);
			break;
		case 0x11f:	//	translated from VK_F6
			PostMessage(hWnd, WM_COMMAND, MAKELONG(ID_REC_DATAREC, 1), 0);
			break;
		case 0x120:	//	translated from VK_F7
			PostMessage(hWnd, WM_COMMAND, MAKELONG(ID_PLAY_DATAREC, 1), 0);
			break;
		case 0x121:	//	translated from VK_F8
			SendMessage(hWnd, WM_COMMAND, MAKELONG(ID_FAST_FORWARD_DATAREC, 1), 0);
			break;
		case 0x122:	//	translated from VK_F9
			PostMessage(hWnd, WM_COMMAND, MAKELONG(ID_OPEN_FD1, 1), 0);
			break;
		case 0x123:	//	translated from VK_F10
			PostMessage(hWnd, WM_COMMAND, MAKELONG(ID_OPEN_FD2, 1), 0);
			break;
	}
}

/// translate
int EMU::translate_global_key(int code)
{
	if (VK_F1 <= code && code <= VK_F24) {
		code += (0x011a - VK_F1);	// 0x11a - 0x132
	}
	return code;
}

