/** @file winutils.cpp

	Skelton for retropc emulator

	@author Sasaji
	@date   2012.08.01 -

	@brief [ win32 utils ]
*/

#include "winutils.h"

extern EMU *emu;

HMENU get_sub_menu(HMENU hMenu, int id) {
	HMENU hSubMenu = NULL;
	int count = GetMenuItemCount(hMenu);
	for(int n=count-1; n>=0; n--) {
		hSubMenu = GetSubMenu(hMenu,n);
		if (!hSubMenu) continue;

		MENUITEMINFO mii;
		memset(&mii, 0, sizeof(mii));
		mii.cbSize = sizeof(mii);
		mii.fMask = MIIM_ID;
		if (!GetMenuItemInfo(hSubMenu, id, FALSE, &mii)) {
			hSubMenu = NULL;
			continue;
		}
		break;
	}
	return hSubMenu;
}

void get_long_full_path_name(_TCHAR* src, _TCHAR* dst)
{
	_TCHAR tmp[_MAX_PATH];

	if(GetFullPathName(src, _MAX_PATH, tmp, NULL) == 0) {
		// not restrict
		_tcsncpy(tmp, src, _MAX_PATH);
		_tcsncpy(dst, tmp, _MAX_PATH);
	}
	else if(GetLongPathName(tmp, dst, _MAX_PATH) == 0) {
		_tcsncpy(dst, tmp, _MAX_PATH);
	}
}

#if 0
_TCHAR* get_parent_dir(_TCHAR* file)
{
	static _TCHAR path[_MAX_PATH];

	_tcscpy(path, file);
	int pt = (int)_tcslen(path);
	while(pt >= 0 && path[pt] != _T('\\')) {
		pt--;
	}
	path[pt + 1] = _T('\0');
	return path;
}
#endif

/// file dialog
_TCHAR* get_open_file_name(HWND hWnd, _TCHAR* filter, _TCHAR* title, _TCHAR* dir, _TCHAR* ext, bool save, DWORD* flags)
{
	FileBox fbox(hWnd);
	static _TCHAR path[_MAX_PATH];

	if (emu) emu->set_pause(1, true);

	bool rc = fbox.Show(filter, title, dir, ext, save, path);

	if (flags != NULL) *flags = fbox.GetFlags();

	if (emu) emu->set_pause(1, false);

	if (rc) {
		return path;
	}
	return NULL;
}

_TCHAR* get_open_file_name(HWND hWnd, _TCHAR* filter, _TCHAR* title, _TCHAR* dir, bool save)
{
	return get_open_file_name(hWnd, filter, title, dir, NULL, save, NULL);
}

////////////////////
#ifdef USE_CART
void open_cart_dialog(HWND hWnd)
{
	_TCHAR* path = get_open_file_name(
		hWnd,
#ifdef _X1TWIN
		_T("Supported Files (*.pce)\0*.pce\0All Files (*.*)\0*.*\0\0"),
		_T("HuCARD"),
#else
		_T("Supported Files (*.rom;*.bin)\0*.rom;*.bin\0All Files (*.*)\0*.*\0\0"),
		_T("Game Cartridge"),
#endif
		config.initial_cart_path
	);
	if(path) {
		UPDATE_HISTORY(path, config.recent_cart_path);
		emu->open_cart(path);
	}
}
#endif

#ifdef USE_FD1
bool open_disk_dialog(HWND hWnd, int drv)
{
	DWORD flags;
	_TCHAR title[128];
	_stprintf(title, _T("Open Floppy Disk %d"), drv);

	_TCHAR* path = get_open_file_name(
		hWnd,
#ifndef _BML3MK5
		_T("Supported Files (*.d88;*.d77;*.td0;*.imd;*.dsk;*.fdi;*.hdm;*.tfd;*.xdf;*.2d;*.sf7)\0*.d88;*.d77;*.td0;*.imd;*.dsk;*.fdi;*.hdm;*.tfd;*.xdf;*.2d;*.sf7\0All Files (*.*)\0*.*\0\0"),
#else
		_T("Supported Files (*.d88;*.td0;*.imd;*.dsk;*.fdi;*.hdm;*.tfd;*.xdf;*.2d;*.sf7)\0*.d88;*.td0;*.imd;*.dsk;*.fdi;*.hdm;*.tfd;*.xdf;*.2d;*.sf7\0All Files (*.*)\0*.*\0\0"),
#endif
		title,
		config.initial_disk_path,
		_T("d88"),
		false,
		&flags
	);
	if(path) {
		return open_disk(drv, path, 0, flags, true);
	} else {
		return false;
	}
}

bool open_disk(int drv, _TCHAR* path, int bank, uint32 flags, bool multiopen)
{
	get_long_full_path_name(path, path);
	return emu->open_disk_by_bank_num(drv, path, bank, flags, multiopen);
}

void close_disk(int drv)
{
	emu->close_disk(drv);
}

void set_disk_file_menu(HMENU hMenu, UINT uItem, int drv)
{
	MENUITEMINFO minfo;
	_TCHAR str[_MAX_PATH];

	minfo.cbSize = sizeof(MENUITEMINFO);
	minfo.fMask = MIIM_TYPE;
	minfo.fType = MFT_STRING;

	bool ins = emu->disk_inserted(drv);

	if (ins) {
		_tcscpy(str,_T("Open ["));
		_tcscat(str,UTILITY::trim_center(config.opened_disk_path[drv].path, TRIM_STRING_SIZE));
		if (config.opened_disk_path[drv].num > 0) {
			_stprintf(&str[_tcslen(str)], _T(" : %d"), config.opened_disk_path[drv].num + 1);
		}
		_tcscat(str,_T("]"));
	} else {
		_tcscpy(str,_T("Open..."));
	}

	minfo.dwTypeData = str;
	minfo.cch = (UINT)_tcslen(str);

	SetMenuItemInfo(hMenu, uItem, FALSE, &minfo);
	CheckMenuItem(hMenu, uItem, ins ? MF_CHECKED : MF_UNCHECKED);
}

#ifdef _BML3MK5
void set_disk_side_menu(HMENU hMenu, UINT uItem, int drv)
{
	if (emu) {
		if (config.fdd_type != 1) {
			EnableMenuItem(hMenu, uItem, MF_GRAYED);
		} else {
			MENUITEMINFO minfo;
			_TCHAR str[40];

			minfo.cbSize = sizeof(MENUITEMINFO);
			minfo.fMask = MIIM_TYPE;
			minfo.fType = MFT_STRING;

			int side = emu->get_disk_side(drv);

			if (side) {
				_tcscpy(str,_T("Change side to A"));
			} else {
				_tcscpy(str,_T("Change side to B"));
			}
			minfo.dwTypeData = str;
			minfo.cch = (UINT)_tcslen(str);

			SetMenuItemInfo(hMenu, uItem, FALSE, &minfo);
			EnableMenuItem(hMenu, uItem, MF_ENABLED);
		}
	}
}
#endif
#endif	// USE_FD1

#ifdef USE_QUICKDISK
void open_quickdisk_dialog(HWND hWnd)
{
	_TCHAR* path = get_open_file_name(
		hWnd,
		_T("Supported Files (*.mzt)\0*.mzt\0All Files (*.*)\0*.*\0\0"),
		_T("Quick Disk"),
		config.initial_quickdisk_path
	);
	if(path) {
		UPDATE_HISTORY(path, config.recent_quickdisk_path);
		emu->open_quickdisk(path);
	}
}
#endif

#ifdef USE_DATAREC
void open_play_datarec_dialog(HWND hWnd)
{
	_TCHAR* path = get_open_file_name(
		hWnd,
#if defined(DATAREC_PC8801)
		_T("Supported Files (*.cas;*.cmt;*.t88)\0*.cas;*.cmt;*.t88\0All Files (*.*)\0*.*\0\0"),
#elif defined(DATAREC_BINARY_ONLY)
		_T("Supported Files (*.cas;*.cmt)\0*.cas;*.cmt\0All Files (*.*)\0*.*\0\0"),
#elif defined(DATAREC_TAP)
		_T("Supported Files (*.wav;*.cas;*.tap)\0*.wav;*.cas;*.tap\0All Files (*.*)\0*.*\0\0"),
#elif defined(DATAREC_MZT)
		_T("Supported Files (*.wav;*.cas;*.mzt;*.m12)\0*.wav;*.cas;*.mzt;*.m12\0All Files (*.*)\0*.*\0\0"),
#elif defined(_BML3MK5)
		_T("Supported Files (*.l3;*.l3b;*.l3c;*.wav;*.t9x)\0*.l3;*.l3b;*.l3c;*.wav;*.t9x\0All Files (*.*)\0*.*\0\0"),
#else
		_T("Supported Files (*.wav;*.cas)\0*.wav;*.cas\0All Files (*.*)\0*.*\0\0"),
#endif
		_T("Play Data Recorder Tape"),
		config.initial_datarec_path,
		_T("l3"),
		false,
		NULL
	);
	if(path) {
		load_datarec(path);
	}
}

void open_record_datarec_dialog(HWND hWnd)
{
	_TCHAR* path = get_open_file_name(
		hWnd,
#if defined(DATAREC_PC8801)
		_T("Supported Files (*.cas;*.cmt)\0*.cas;*.cmt\0All Files (*.*)\0*.*\0\0"),
#elif defined(DATAREC_BINARY_ONLY)
		_T("Supported Files (*.cas;*.cmt)\0*.cas;*.cmt\0All Files (*.*)\0*.*\0\0"),
#elif defined(DATAREC_TAP)
		_T("Supported Files (*.wav;*.cas)\0*.wav;*.cas\0All Files (*.*)\0*.*\0\0"),
#elif defined(DATAREC_MZT)
		_T("Supported Files (*.wav;*.cas)\0*.wav;*.cas\0All Files (*.*)\0*.*\0\0"),
#elif defined(_BML3MK5)
		_T("L3 File (*.l3)\0*.l3\0L3B File (*.l3b)\0*.l3b\0L3C File (*.l3c)\0*.l3c\0Wave File (*.wav)\0*.wav\0T9X File (*.t9x)\0*.t9x\0All Files (*.*)\0*.*\0\0"),
#else
		_T("Supported Files (*.wav;*.cas)\0*.wav;*.cas\0All Files (*.*)\0*.*\0\0"),
#endif
		_T("Record Data Recorder Tape"),
		config.initial_datarec_path,
		_T("l3"),
		true,
		NULL
	);
	if(path) {
		save_datarec(path);
	}
}

void set_datarec_file_menu(HMENU hMenu, UINT uItem)
{
	MENUITEMINFO minfo;
	_TCHAR str[_MAX_PATH];

	minfo.cbSize = sizeof(MENUITEMINFO);
	minfo.fMask = MIIM_TYPE;
	minfo.fType = MFT_STRING;

	if (emu->datarec_opened(true)) {
		_tcscpy(str,_T("Play ["));
		_tcscat(str, UTILITY::trim_center(config.opened_datarec_path.path, TRIM_STRING_SIZE));
		_tcscat(str,_T("]"));
	} else {
		_tcscpy(str,_T("Play..."));
	}
	minfo.dwTypeData = str;
	minfo.cch = (UINT)_tcslen(str);
	SetMenuItemInfo(hMenu, uItem, FALSE, &minfo);

	if (emu->datarec_opened(false)) {
		_tcscpy(str,_T("Rec ["));
		_tcscat(str, UTILITY::trim_center(config.opened_datarec_path.path, TRIM_STRING_SIZE));
		_tcscat(str,_T("]"));
	} else {
		_tcscpy(str,_T("Rec..."));
	}
	minfo.dwTypeData = str;
	minfo.cch = (UINT)_tcslen(str);
	SetMenuItemInfo(hMenu, uItem + 1, FALSE, &minfo);

//	CheckMenuRadioItem(hMenu, uItem, uItem + 1, uItem + opened_tape_type - 1, MF_BYCOMMAND);
	CheckMenuItem(hMenu, uItem, emu->datarec_opened(true) ? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(hMenu, uItem + 1, emu->datarec_opened(false) ? MF_CHECKED : MF_UNCHECKED);
}

bool load_datarec(_TCHAR *file_path)
{
	get_long_full_path_name(file_path, file_path);
	return emu->play_datarec(file_path);
}

bool save_datarec(_TCHAR *file_path)
{
	get_long_full_path_name(file_path, file_path);
	return emu->rec_datarec(file_path);
}

void close_datarec()
{
//	opened_tape_type = 0;
	emu->close_datarec();
}
#endif

#ifdef USE_MEDIA
void open_media_dialog(HWND hWnd)
{
	_TCHAR* path = get_open_file_name(
		hWnd,
		_T("Supported Files (*.m3u)\0*.m3u\0All Files (*.*)\0*.*\0\0"),
		_T("Sound Cassette Tape"),
		config.initial_media_path
	);
	if(path) {
		UPDATE_HISTORY(path, config.recent_media_path);
		emu->open_media(path);
	}
}
#endif

#ifdef USE_BINARY_FILE1
void open_binary_dialog(HWND hWnd, int drv, bool load)
{
	_TCHAR* path = get_open_file_name(
		hWnd,
		_T("Supported Files (*.ram;*.bin)\0*.ram;*.bin\0All Files (*.*)\0*.*\0\0"),
#if defined(_PASOPIA) || defined(_PASOPIA7)
		_T("RAM Pack Cartridge"),
#else
		_T("Memory Dump"),
#endif
		config.initial_binary_path
	);
	if(path) {
		UPDATE_HISTORY(path, config.recent_binary_path[drv]);
		if(load) {
			emu->load_binary(drv, path);
		}
		else {
			emu->save_binary(drv, path);
		}
	}
}
#endif

#ifdef USE_PRINTER
void open_printer_dialog(HWND hWnd, int dev)
{
	_TCHAR* path = get_open_file_name(
		hWnd,
		_T("Supported Files (*.lpt)\0*.lpt\0All Files (*.*)\0*.*\0\0"),
		_T("Print Data"),
		config.initial_printer_path,
		_T("lpt"),
		TRUE,
		NULL
		);
	if(path) {
		save_printer(dev, path);
	}
}

bool save_printer(int dev, _TCHAR *file_path)
{
	get_long_full_path_name(file_path, file_path);
	return emu->save_printer(dev, file_path);
}
#endif

#ifdef USE_AUTO_KEY
void open_autokey_dialog(HWND hWnd)
{
	_TCHAR* path = get_open_file_name(
		hWnd,
		_T("Supported Files (*.txt;*.bas;*.lpt)\0*.txt;*.bas;*.lpt\0All Files (*.*)\0*.*\0\0"),
		_T("Open Auto key file"),
		config.initial_autokey_path,
		_T("txt"),
		false,
		NULL
		);
	if(path) {
		load_autokey(path);
	}
}

bool load_autokey(_TCHAR *file_path)
{
	get_long_full_path_name(file_path, file_path);
	return emu->start_auto_key(file_path);
}
#endif

#ifdef _BML3MK5
void open_save_state_dialog(HWND hWnd)
{
	memset(config.saved_state_path, 0, sizeof(config.saved_state_path));
	_TCHAR* path = get_open_file_name(
		hWnd,
		_T("Supported Files (*.l3r)\0*.l3r\0All Files (*.*)\0*.*\0\0"),
		_T("Save Status Data"),
		config.initial_state_path,
		_T("l3r"),
		true,
		NULL
		);
	if(path) {
		save_state(path);
	}
}

void open_load_state_dialog(HWND hWnd)
{
	_TCHAR* path = get_open_file_name(
		hWnd,
		_T("Supported Files (*.l3r)\0*.l3r\0All Files (*.*)\0*.*\0\0"),
		_T("Load Status Data"),
		config.initial_state_path,
		_T("l3r"),
		false,
		NULL
		);
	if(path) {
		load_state(path);
	}
}

bool save_state(_TCHAR *file_path)
{
	get_long_full_path_name(file_path, file_path);
	return emu->save_state(file_path);
}

bool load_state(_TCHAR *file_path)
{
	get_long_full_path_name(file_path, file_path);
	return emu->load_state(file_path);
}

#ifdef USE_KEY_RECORD
void open_play_reckey_dialog(HWND hWnd)
{
	_TCHAR* path = get_open_file_name(
		hWnd,
		_T("Supported Files (*.l3k)\0*.l3k\0All Files (*.*)\0*.*\0\0"),
		_T("Play recorded keys"),
		config.initial_state_path,
		_T("l3k"),
		false,
		NULL
		);
	if(path) {
		load_reckey(path);
	}
}

void open_record_reckey_dialog(HWND hWnd)
{
	_TCHAR* path = get_open_file_name(
		hWnd,
		_T("Supported Files (*.l3k)\0*.l3k\0All Files (*.*)\0*.*\0\0"),
		_T("Record input keys"),
		config.initial_state_path,
		_T("l3k"),
		true,
		NULL
		);
	if(path) {
		save_reckey(path);
	}
}

bool save_reckey(_TCHAR *file_path)
{
	get_long_full_path_name(file_path, file_path);
	return emu->record_reckey(file_path);
}

bool load_reckey(_TCHAR *file_path)
{
	get_long_full_path_name(file_path, file_path);
	return emu->play_reckey(file_path);
}

#endif
#endif	// _BML3MK5

/**
 *	open dropped file
 */
void open_dropped_file(HWND hWnd, WPARAM wParam)
{
	_TCHAR file_path[_MAX_PATH];

    HDROP hDrop=(HDROP)wParam;
	// get first filename
	DragQueryFile(hDrop, 0, file_path, _MAX_PATH);
    DragFinish(hDrop);

	if (_tcslen(file_path) == 0) {
		return;
	}
	int rc = check_supported_file(file_path);
	switch(rc) {
	case 1:
		// maybe tape image file
		load_datarec(file_path);
		break;
#ifdef USE_FD1
	case 2:
		// maybe disk image file
		open_disk(0, file_path, 0, 0, true);
		break;
#endif
#ifdef _BML3MK5
	case 3:
		// maybe state file
		load_state(file_path);
		break;
	case 4:
		// maybe text file (for autokey)
		load_autokey(file_path);
		break;
	case 6:
		// maybe record key file
		load_reckey(file_path);
		break;
#endif
	default:
		break;
	}
}

/**
 *	check supported file
 */
int check_supported_file(_TCHAR *file_path)
{
	bool rc;

	rc = UTILITY::check_file_extension(file_path, _T(".l3"))
	  || UTILITY::check_file_extension(file_path, _T(".l3b"))
	  || UTILITY::check_file_extension(file_path, _T(".l3c"))
	  || UTILITY::check_file_extension(file_path, _T(".wav"))
	  || UTILITY::check_file_extension(file_path, _T(".t9x"));
	if (rc) {
		// tape image
		return 1;
	}

#ifdef USE_FD1
	rc = UTILITY::check_file_extension(file_path, _T(".d88"))
	  || UTILITY::check_file_extension(file_path, _T(".td0"))
	  || UTILITY::check_file_extension(file_path, _T(".imd"))
	  || UTILITY::check_file_extension(file_path, _T(".dsk"))
	  || UTILITY::check_file_extension(file_path, _T(".fdi"))
	  || UTILITY::check_file_extension(file_path, _T(".hdm"))
	  || UTILITY::check_file_extension(file_path, _T(".tfd"))
	  || UTILITY::check_file_extension(file_path, _T(".xdf"))
	  || UTILITY::check_file_extension(file_path, _T(".2d"))
	  || UTILITY::check_file_extension(file_path, _T(".sf7"));
	if (rc) {
		// disk image
		return 2;
	}
#endif

#ifdef _BML3MK5
	rc = UTILITY::check_file_extension(file_path, _T(".l3r"));
	if (rc) {
		// maybe state file
		return 3;
	}

	rc = UTILITY::check_file_extension(file_path, _T(".txt"))
	  || UTILITY::check_file_extension(file_path, _T(".bas"))
	  || UTILITY::check_file_extension(file_path, _T(".lpt"));
	if (rc) {
		// maybe text file (for autokey)
		return 4;
	}

	rc = UTILITY::check_file_extension(file_path, _T(".ini"));
	if (rc) {
		// maybe ini file
		return 5;
	}

	rc = UTILITY::check_file_extension(file_path, _T(".l3k"));
	if (rc) {
		// maybe record key file
		return 6;
	}
#endif

	return 0;
}
