/** @file winutils.h

	Skelton for retropc emulator

	@author Sasaji
	@date   2012.08.01 -

	@brief [ win32 utils ]
*/

#ifndef _WINUTILS_H_
#define _WINUTILS_H_

#include <windows.h>
#include <stdio.h>
#include "emu.h"
#include "gui/filebox.h"

#define TRIM_STRING_SIZE	72

#define UPDATE_RECENT_MENU(ID_RECENT, recent_path) { \
	bool flag = false; \
	HMENU hSubMenu = get_sub_menu(hMenu, ID_RECENT); \
	_TCHAR str[_MAX_PATH]; \
	if (hSubMenu) { \
		while(DeleteMenu(hSubMenu, 0, MF_BYPOSITION) != 0) {} \
		for(int i = 0; i < MAX_HISTORY; i++) { \
			if(!_tcscmp(recent_path[i].path, _T(""))) break; \
			_tcscpy(str,UTILITY::trim_center(recent_path[i].path, TRIM_STRING_SIZE)); \
			if (recent_path[i].num > 0) { \
				_stprintf(&str[_tcslen(str)], _T(" : %d"), recent_path[i].num + 1); \
			} \
			AppendMenu(hSubMenu, MF_STRING, ID_RECENT + i, str); \
			flag = true; \
		} \
		if(!flag) { \
			AppendMenu(hSubMenu, MF_GRAYED | MF_STRING, ID_RECENT, _T("None")); \
		} \
	} \
}

#define UPDATE_MULTIVOLUME_MENU(drv, ID_D88_FILE_PATH, ID_SELECT_BANK) { \
	bool flag = false; \
	HMENU hSubMenu = get_sub_menu(hMenu, ID_SELECT_BANK); \
	if (hSubMenu) { \
		while(DeleteMenu(hSubMenu, 0, MF_BYPOSITION) != 0) {} \
		d88_file_t *d88_file = emu->get_d88_bank(drv); \
		if(d88_file->bank_nums >= 1) { \
			for(int i = 0; i < d88_file->bank_nums; i++) { \
				_TCHAR tmp[32]; \
				_TCHAR *name = d88_file->bank[i].name; \
				_stprintf(tmp, _T("%2d: %s"), i + 1, name[0] != _T('\0') ? name : _T("(no label)")); \
				AppendMenu(hSubMenu \
					, MF_STRING | (d88_file->bank_nums <= 1 ? MF_GRAYED : 0) | (i == d88_file->cur_bank ? MF_CHECKED : 0) \
					, ID_SELECT_BANK + i, tmp); \
			} \
			flag = true; \
		} \
		if(!flag) { \
			AppendMenu(hSubMenu, MF_GRAYED | MF_STRING, ID_SELECT_BANK, _T("None")); \
		} \
	} \
}

// prototype
HMENU get_sub_menu(HMENU hMenu, int id);

void    get_long_full_path_name(_TCHAR* src, _TCHAR* dst);
//_TCHAR* get_parent_dir(_TCHAR* file);

_TCHAR* get_open_file_name(HWND hWnd, _TCHAR* filter, _TCHAR* title, _TCHAR* dir, _TCHAR* ext = NULL, bool save = false, DWORD* flags = NULL);
_TCHAR* get_open_file_name(HWND hWnd, _TCHAR* filter, _TCHAR* title, _TCHAR* dir, bool save);


// dialog
#ifdef USE_CART
void open_cart_dialog(HWND hWnd);
#endif
#ifdef USE_FD1
bool open_disk_dialog(HWND hWnd, int drv);
bool open_disk(int drv, _TCHAR* filename, int bank, uint32 flags, bool multiopen);
void close_disk(int drv);
void set_disk_file_menu(HMENU hMenu, UINT uItem, int drv);
#ifdef _BML3MK5
void set_disk_side_menu(HMENU hMenu, UINT uItem, int drv);
#endif
#endif
#ifdef USE_QUICKDISK
void open_quickdisk_dialog(HWND hWnd);
#endif
#ifdef USE_DATAREC
void open_play_datarec_dialog(HWND hWnd);
void open_record_datarec_dialog(HWND hWnd);
void set_datarec_file_menu(HMENU hMenu, UINT uItem);
bool load_datarec(_TCHAR *file_path);
bool save_datarec(_TCHAR *file_path);
void close_datarec();
#endif
#ifdef USE_MEDIA
void open_media_dialog(HWND hWnd);
#endif
#ifdef USE_BINARY_FILE1
void open_binary_dialog(HWND hWnd, int drv, bool load);
#endif
#ifdef USE_PRINTER
void open_printer_dialog(HWND hWnd, int drv);
bool save_printer(int dev, _TCHAR *file_path);
#endif
#ifdef USE_AUTO_KEY
void open_autokey_dialog(HWND hWnd);
bool load_autokey(_TCHAR *file_path);
#endif
#ifdef _BML3MK5
void open_save_state_dialog(HWND hWnd);
void open_load_state_dialog(HWND hWnd);
bool save_state(_TCHAR *file_path);
bool load_state(_TCHAR *file_path);
#ifdef USE_KEY_RECORD
void open_play_reckey_dialog(HWND hWnd);
void open_record_reckey_dialog(HWND hWnd);
bool load_reckey(_TCHAR *file_path);
bool save_reckey(_TCHAR *file_path);
#endif
#endif
void open_dropped_file(HWND hWnd, WPARAM wParam);
int check_supported_file(_TCHAR *file_path);

#endif
